// Copyright  Michael Monagan  2020
//
// Frontend to calling MTSHL for sparse Hensel lifting
//
// This file was created by making a copy of factor4.cilk stripping out the
// Cilk directives spawn, sync, cilk using  s/spawn//  s/sync//  s/cilk / /
// Also modifying the command line inputs to not have a number of cores
// I also had to recode the timer below as the Cilk version is different.


#include <stdio.h>
#include <time.h>
#include <stdlib.h>

// Timer
double realtime()
{
        clock_t st;
        double ST;
        st = clock(); ST = (1.0*st)/1000000;
        return ST;
}

// Timer 
double realtimems() {
        clock_t st;
        double ST;
        st = clock(); ST = (1.0*st)/1000;
        return ST;
}


#define LONG long long int
#define ULONG unsigned long long int
ULONG seed;
ULONG mult;
LONG rand64s(LONG p);
int abs32s(int x);

int * arrayints( int n );
LONG * array( int n );
LONG ** arrayofptrs( int n );
int ** arrayptrsints( int n );
char * arrayofbytes( int n );
void freemem( void *p );
void srand(unsigned int seed);
int rand(void);
void vecprint32s( int *A, int n );

/******************************************************************************************/
/*  Zp utilities                                                                          */
/******************************************************************************************/

#define max32s max32smain
#define min32s min32smain
#define add64s add64smain
#define sub64s sub64smain
#define mul64s mul64smain
#define neg64s neg64smain

inline int max32s(int a, int b) { if(a>b) return a; else return b; }
inline int min32s(int a, int b) { if(a<b) return a; else return b; }
inline LONG add64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
inline LONG sub64s(LONG a, LONG b, LONG p) { LONG t; t = a-b; t += (t>>63) & p; return t; }
inline LONG neg64s(LONG a, LONG p) { LONG t; t = -a; t += (t>>63) & p; return t; }
inline LONG mul64s(LONG a, LONG b, LONG p) {
        LONG q, r;
        __asm__ __volatile__(           \
        "       mulq    %%rdx           \n\t" \
        "       divq    %4              \n\t" \
        : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
        return r;
}

LONG modinv64s( LONG c, LONG p );
LONG powmod64s( LONG a, LONG n, LONG p );
LONG prevFourierprime64s( LONG x, int k );
LONG prevPrime64s( LONG x);

typedef struct P_INT P_INT;
struct P_INT {
        LONG p;        /* prime */
        LONG s;        /* shift */
        LONG d;        /* normalized */
        LONG v;        /* reciprocal */
};

#define modinit modinitmain
#define mulmodinv mulmodinvmain

/* initialize mod p struct */
inline P_INT modinit(LONG p)
{
        P_INT m;

        /* most significant bit location 63..0 */
        __asm__ __volatile__("bsrq %0, %0" : "=r"(m.s) : "0"(p));
        m.p = p;
        m.s = 64-m.s-1;
        m.d = p << m.s;

        /* reciprocal: (2^128-1)/n - 2^64 */
        /* [(2^64-n-1)*2^64 + (2^64-1)]/n */
        __asm__ __volatile__(\
        "       movq    %1, %%rdx       \n\t" \
        "       addq    $1, %%rdx       \n\t" \
        "       negq    %%rdx           \n\t" \
        "       movq    $-1, %%rax      \n\t" \
        "       divq    %1              \n\t" \
                : "=a"(m.v) : "r"(m.d) : "%rdx");
        return m;
}

inline LONG mulmodinv(LONG a, LONG b, P_INT m)
{
        LONG q, r;
        /* 0: compute  a*b = u1:u0, lshift by s, r9:r8 = (u1+1):u0 */
        /* 1: compute u1*v = q1:q0, q1:q0 += (u1+1):u0, in rdx:rax */
        /* 2: compute    r = u0-q1*d in r8 */
        /* 3: if r > q0 then q1 = q1-1, r = r+d (possible) */
        /* 4: if r >= d then q1 = q1+1, r = r-d (unlikely) */
        __asm__ __volatile__(\
        "0:     mulq    %%rdx                   \n\t" \
        "       shldq   %%cl,  %%rax, %%rdx     \n\t" \
        "       shlq    %%cl,  %%rax            \n\t" \
        "       leaq    1(%%rdx), %%r9          \n\t" \
        "       movq    %%rax, %%r8             \n\t" \
        "1:     movq    %%rdx, %%rax            \n\t" \
        "       mulq    %5                      \n\t" \
        "       addq    %%r8,  %%rax            \n\t" \
        "       adcq    %%r9,  %%rdx            \n\t" \
        "2:     imulq   %4,    %%rdx            \n\t" \
        "       subq    %%rdx, %%r8             \n\t" \
        "3:     cmpq    %%rax, %%r8             \n\t" \
        "       jbe     4f                      \n\t" \
        "       addq    %4,    %%r8             \n\t" \
        "       decq    %%rax                   \n\t" \
        "4:     cmpq    %4,    %%r8             \n\t" \
        "       jae     5f                      \n\t" \
        "       jmp     6f                      \n\t" \
        "5:     subq    %4,    %%r8             \n\t" \
        "       incq    %%rax                   \n\t" \
        "6:     shrq    %%cl,  %%r8             \n\t" \
        "       movq    %%r8,  %%rdx            \n\t" \
        : "=a"(q), "=d"(r)
        :  "0"(a),  "1"(b), "r"(m.d), "r"(m.v), "c"(m.s)
        : "%r8", "%r9", "%r10");
        /* q = quotient */
        return r;
}

#define UINT64 unsigned long long

typedef struct {
	UINT64 s;	/* shift */
	UINT64 v;	/* reciprocal of d */
	UINT64 d0;	/* divisor shifted up */
	UINT64 d1;
} recint;

recint recip1(UINT64  p);
UINT64 mulrec64(UINT64  a, UINT64  b, recint  v);

/******************************************************************************************/
/* Polynomial utilities                                                                   */
/******************************************************************************************/

void VECADDINP64s( LONG *A, LONG *B, int n, LONG p );
void VECMUL64s( LONG *A, LONG *B, LONG *C, int n, LONG p );
void VECMULINP64s( LONG *A, LONG *B, int n, LONG p );
void VECSCAMUL64s( LONG x, LONG *A, int n, LONG p );
void VECPOWMOD64s( LONG *A, LONG m, int n, LONG p );
void VECPOWERS64s( LONG alpha, LONG *A, int n, LONG p );

void vecfill64s( LONG x, LONG *A, int n );
void vecprint64s( LONG *A, int n );
void vecprint64u( ULONG *A, int n );
void vecpow64s( int n, LONG *A, LONG alpha, LONG p );
void sort2vec64s(LONG *D, LONG *C, int n );


int poladd64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG p);
int polsub64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG p);
void polcopy64s( LONG *A, int d, LONG *B );
void polprint64s( LONG *A, int d );
void polnsparseprint64s( LONG *C, LONG *X, int t, int n );
void polnsparseprint64ZG( LONG *C, LONG *X, int* SC, int t, int n,int bsize);
void pol2sort64s(LONG* C, LONG* X, int n);
LONG pol2simp64s( LONG *C, LONG *X, int n, LONG p );
LONG pol2simp64G( LONG *C, LONG *X, int* SC, int n, int m );
LONG pol2simp64Gx( LONG **C, LONG *X, int* SC, int n, int m );
int polmul64s( LONG * A, LONG * B, LONG * C, int da, int db, LONG p);
int poldiv64s( LONG * A, LONG * B, int da, int db, LONG p );
void polscamul64s(LONG x, LONG* A, int d, LONG p);
void polscamulrec64s( LONG x, LONG *A, int d, LONG p, recint P );
void monic64s(LONG* A, int d, LONG p);
LONG poleval64s(LONG *a, int d, LONG x, LONG p);
void polnsparsevecdeg64s( LONG *X, int t, int n, int *D );
int polnsparsedeg64s( LONG *X, int t, int n, int xi );
LONG polnsparseeval64s(LONG *C, LONG *X, int t, LONG *A, int n, LONG p );
int poln2nm1sparseeval64s( LONG *C, LONG *X, int t, LONG *A, int n, LONG *B, LONG *Y, LONG p );
void randpoly64s( int n, int *DEG, int t, LONG *A, LONG *X, LONG p );

void randpolymonic64s( int n, int *DEG, int t, LONG *A, LONG *X, LONG p );

void randpolyNOTmonic64Gx( int n, int *DEG, int D, int t, LONG **A, int*SA, LONG *X, int m, int mu, int symmetric);
void randpolyNOTmonic64s( int n, int *DEG, int t, LONG *A, LONG *X, LONG p );
int polgcd64s( LONG * A, LONG * B, int da, int db, LONG p );
int polsparsemul64s( LONG *A, LONG *X, int m, LONG *B, LONG *Y, int n, LONG p , LONG *C, LONG *Z );
int polsparseadd64s( LONG *A, LONG *X, int m, LONG *B, LONG *Y, int n, LONG p , LONG *C, LONG *Z );
void polmod64Gx( int d, int n, LONG *A, LONG **B, int* SB, LONG p,recint P, LONG*X );
LONG evalhorner( LONG *f, LONG d, LONG x, recint P, LONG p );
void evalf0tod( LONG *f, int df, int d, LONG *Y, long p );
int interp0tod64s( LONG *Y, int d, LONG *V, LONG *F, LONG p );
void newt2stand( LONG *V, int d, LONG *B, recint P, LONG p );
int newtoncofs( LONG *U, int d, LONG *minv, LONG *V, recint P, LONG p );

int poln2nm1Seval64s( LONG *C, LONG *X, int t, int n, LONG *f, int d, int xshift, LONG p )
{
    // addup coefficients in C of same degree in x1 (in X) and put in f
    int i,j;
    LONG c,x;
    //int numbits;
    //numbits = 63/n;
    //xshift = numbits*(n-1);
    for( i=0; i<=d; i++ ) f[i] = 0;
    for( i=0; i<t; i=j ) {
         c = C[i]; // coefficient
         x = X[i]; // monomial
         x = x >> xshift;
         for( j=i+1; j<t && (X[j] >> xshift) == x; j++ ) c = add64s(c,C[j],p);
         f[x] = add64s(c,f[x],p); // supposed to be sorted on x1 but just in case
    }
    for( ; d>=0 && f[d]==0; d-- );
    return d;
}

void poln2nm2Seval64s( LONG *C, LONG *X, int t, int n, LONG *fc, int dy, int xshift, LONG p )
{
    // addup coefficients in C of same degree in x1,x2 (in X) and put in fc
    int i, numbits;
    LONG mask, xcoeff, ycoeff, x;
    numbits = 63/n;
    mask = (1LL << numbits)-1;
    for( i=0; i<t; i++ )
    {
        ycoeff = (X[i] >> xshift) & mask;
        xcoeff = X[i] >> (xshift+numbits);
        x = (dy+1)*xcoeff+ycoeff;
        fc[x] = add64s(C[i],fc[x],p); // monomial
    }
}

int pol2compress64s(LONG* C, LONG* X, int dx, int dy)
{
    // convert polynomial in x1,x2 into bivariate sparse format (15 bits each variable)
    int i, j;
    for (i=0, j=0; i < (dx+1)*(dy+1); i++)
    {
        if (C[i] != 0)
        {
            C[j] = C[i];
            X[j] = ((i / (dy+1)) << 15) | (i % (dy+1));
            j++;
        }
    }

    return j;
}

int lcoeff264s( LONG *A, LONG *X, int t, int dx, int dy, LONG *C )
{
    int i, d, numbits, xshift;
    LONG mask;
    numbits = 31/2;
    xshift = numbits;
    mask = (1LL << numbits) - 1;

    for( i=0; i<t; i++ )
    {
        if ((X[i] >> xshift) == dx)
        {
            C[X[i] & mask] = A[i];
        }
    }

    d = dy;
    while( d>=0 && C[d]==0 ) d--;
    return d;
}

void getmons(LONG *X, int t, int n, LONG **POW, LONG *GAMMA, LONG p )
{
    int numbits,mask,i,j;
    LONG m,x;
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    for( i=0; i<t; i++ ) {
         m = 1; // evaluation
         x = X[i]; // monomial
         for( j=0; j<n-1; j++ ) { m = mul64s(m,POW[j][x&mask],p); x = x>>numbits; }
         GAMMA[i] = m;
    }
    return;
}

/* a^n mod p assuming 0 <= a < p < 2^31 */
LONG powmodopt64s( LONG a, LONG n, LONG *s, LONG p )
{   int i; LONG r;
    a += (a>>63) & p; // protect from bad input
    if( n==0 ) return 1;
    if( n==1 ) return a;
    for( r=1,i=0; n>0; i++,n=n/2 ) if( n&1 ) r = mul64s(r,s[i],p);
    return r;
}

/* a^n mod p assuming 0 <= a < p < 2^31 */
LONG POWMODOPT64s( LONG a, LONG n, LONG *s, P_INT P )
{   int i; LONG r;
    if( n==0 ) return 1;
    if( n==1 ) return a;
    for( r=1,i=0; n>0; i++,n=n/2 ) if( n&1 ) r = mulmodinv(r,s[i],P);
    return r;
}


LONG POWMOD64s( LONG alpha, LONG y, P_INT P );
void getmonspow(LONG *X, int t, int n, LONG alpha, LONG *GAMMA, LONG *S, int xshift, P_INT P )
{
    int i;
    LONG y,mask;
    mask = ((LONG) 1 << xshift) - 1;
    for( i=0; i<t; i++ ) {
         y = X[i] & mask; // monomial
         //GAMMA[i] = powmod64s(alpha,y,p);
         //GAMMA[i] = powmodopt64s(alpha,y,S,p);
         //GAMMA[i] = POWMOD64s(alpha,y,P);
         GAMMA[i] = POWMODOPT64s(alpha,y,S,P);
    }
    return;
}

int polroots64s( LONG * f, int d, LONG * R, LONG *W, LONG p );
void polSolve64s( LONG *R, int n, LONG *L, LONG *a, LONG *X, LONG p );

void getLambda( LONG *M, int t, LONG *L, LONG p ) {
// compute (x-M[1])(x-M[2])(x-M[3])...(x-M[t]) vi repeated multiplication for t >= 0
    int i,j;
    LONG alpha;
    P_INT P;
    P = modinit(p);
    L[0] = 1;
printf("M[0]=%lld\n",M[0]);
printf("M[1]=%lld\n",M[1]);
    for( i=0; i<t; i++ ) {
        alpha = M[i];
        L[i+1] = 1;
        for( j=i; j>0; j-- ) L[j] = sub64s(L[j-1],mulmodinv(alpha,L[j],P),p);
        L[0] = neg64s(mulmodinv(alpha,L[0],P),p);
    }
    return;
}

LONG ** polarrayenlarge(LONG **F, int t, int n, int d) {
   int i;
   LONG **W, *S;
   W = arrayofptrs(t+n);
   for( i=0; i<t; i++ ) W[i] = F[i];
   S =  array(n*(d+1));
   for( i=0; i<n*(d+1); i++) S[i] = 0;
   for( i=0; i<n; i++ ) W[t+i] = S+i*(d+1);
   freemem(F);
   return W;
}

void VECOPY( LONG *A, LONG *B, int n ) { int i; for( i=0; i<n; i++ ) B[i]=A[i]; return; }
void veccopy64s( LONG *A, int n, LONG *B ) { int i; for( i=0; i<n; i++ ) B[i] = A[i]; return; }

void polgcdext64s( LONG *A, LONG *B, int da, int db,
                   LONG *G, LONG *S, LONG *T, int *dG, int *dS, int *dT, 
                   LONG *W, LONG p );

int poldiosol64s( LONG *A, LONG *B, LONG *C, LONG *S, int da, int db, int dc, int ds, // input
                  LONG *sigma, LONG *tau, int *dsigma, int *dtau, // output
                  LONG *W, /* temporary space */ LONG p );
void randvecp( int n, LONG * A, LONG p );
LONG ** polmake2rec64s( LONG *A, LONG *X, int t );
LONG ** array2( int n, int m );
int ** array2ints( int n, int m );
void polrec2conv64s( LONG *A, LONG *X, int t, LONG ***B, int **DegB );
int polmul2cof64s( LONG **A, LONG **B, int k, LONG *C, LONG *T, int p );
int polrec2mulcof64s( LONG **A, int *DA, LONG **B, int *DB, int k, LONG *C, LONG *T, int p );
void polprint2rec64s( LONG **A );
void polrec2print64s( int *D, LONG **A );
int polsubsca64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG alpha, LONG p, recint P);
int poladdsca64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG alpha, LONG p, recint P);
int polnewton2eval64s( int dy, int *D, LONG **U, LONG *M, LONG beta, LONG *B, LONG p ) ;
void polrec2tayprint64s( int *D, LONG **A, LONG alpha );
int poldiff64s( LONG *A, LONG *X, int t, int n, int j, LONG p );
int polbivareval64s( LONG *A, LONG *X, int na, LONG *W, LONG *B, int dx, int dy, LONG p );
int poltrivareval64s( LONG *A, LONG *X, int na, LONG *beta, LONG *B, LONG *Y, LONG p );
int poltrivar3eval64s( LONG *A, LONG *X, int na, LONG *beta, LONG *B, LONG *Y, LONG p );
int polrec2eval64s( LONG **A, LONG *B, LONG alpha, LONG p );

int HenselLift1( LONG *A, LONG *X, int na, LONG *f, int df, LONG *g, int dg, LONG alpha, LONG p,
                 LONG *SPACE, // temporary space
                 LONG ** Fxy, int *DegF, LONG ** Gxy, int *DegG,
                 LONG ** Fxyvals, LONG ** Gxyvals );

void polbivarimage64s( int dx, int dz, LONG ***F, int **D );
void newton2standard64s( int dy, int *D, LONG **U, LONG *M, int p );
int rec2bivar64s( int dy, int *D, LONG **F, LONG *A, LONG *X );
void evalpowers64s( LONG alpha, int d, LONG *W, LONG p );


//int HenselLift2( LONG *A, LONG *X, int na, LONG *f, LONG *u, int nf, LONG *g, LONG *v, int ng,
//                 LONG alpha, LONG p,
//                 int *DegFz, int *nF, LONG **F, LONG **Y, int *DegGz, int *nG, LONG **G, LONG **Z );

void evalbeta( LONG *A, int d, LONG beta, recint P );
void evalinit( LONG *X, int na, int n, int xi, int xj, LONG **BETA, LONG *B, LONG *Y, LONG p );
int evalnext( LONG *A, LONG *X, LONG *G, int na, LONG *B, LONG *Y, LONG p );
void printsupp( LONG *X, int t, int n );
int ** array2ints(int n, int m);
int polneval1(LONG *A, LONG *X, int t, LONG *ALPHA, LONG *B, LONG *Y, int n, LONG p );
int HornerExpand( int d, LONG alpha, int n, int *N, LONG **F, LONG **Y, LONG *A, LONG *X, LONG *T1, LONG *T2, LONG p );
int MTSHL(int N, LONG *A,LONG *X,int na,LONG *F,LONG *Y,int nf,LONG *G,LONG *Z,int ng,int n,LONG alpha,LONG p);
int  henselcheck(LONG *A, LONG *X, int na, int n,
                 int dFz, LONG **FC, LONG **FX, int *nF,
                 int dGz, LONG **GC, LONG **GX, int *nG,
                 LONG alpha, LONG p );
void evalhens( LONG *A, LONG *X, int t, int n );
LONG getomega64s( LONG p, LONG n );
void FFT64s1( int n, LONG *a, LONG *W, LONG p, recint P );
void FFT64s2( int n, LONG *a, LONG *W, LONG p, recint P );
void MakeW64( int n, LONG w, LONG *W, LONG p );


void getlambda( LONG *M, int t, LONG *L, LONG p );
void treemul( LONG *M, int t, LONG *L, LONG *T, LONG p );


/********************************************************************************************/
int main( int argc, char *argv[] ) {

   extern ULONG seed, mult;
   LONG p;
   recint P;
   int i,NumCores,good;
   LONG ARGS[10];
   LONG *A, *B, *W;
   LONG *a, *b, *c, *g, *s, *t, *sigma, *tau;
   int da, db, dc, ds, dt, dg, dsigma, dtau;
   double ST,ET,TT;

   //srand(time(NULL)); iseed = rand();  DON"T DO THIS!!!
   //iseed = 1; 
   seed = 1;
   mult = 6364136223846793003ll;

   for( i=0; i<10; i++ ) ARGS[i]=0;
   for( i=1; i<argc; i++ ) sscanf(argv[i],"%lld\n", ARGS+i);

   NumCores = 1;
   //if( ARGS[1]!=0 ) NumCores = ARGS[1];
   //printf("NumCores = %d;\n", NumCores);
   for( i=9; i>1; i-- ) ARGS[i] = ARGS[i-1];
   ARGS[1] = 1; // Where Numcores was

if( 0 ) { // test treemul
      double ST,ET;
      int n,okay;
      p = 5; p = (p<<55)+1;
      n = 30000;
      A = array(n);
      W = array(n);
      for( i=0; i<n; i++ ) A[i] = rand64s(p);
      a = array(n+1);
      b = array(n+1);
      ST = realtime();
      getlambda(A,n,a,p);
      ET = realtime();
      if( n<10 ) { printf("Lambda1 := "); polprint64s( a, n ); }
      else printf("n=%d  getlambda=%8.2f\n",n,ET-ST);
      ST = ET;
      treemul(A,n,b,W,p);
      ET = realtime();
      if( n<10 ) { printf("Lambda2 := "); polprint64s( b, n ); }
      else printf("n=%d  treemul=%8.2f\n",n,ET-ST);
      for( okay=1,i=0; i<=n && okay; i++ ) if( a[i]!=b[i] ) okay = 0;
      if( okay ) printf("Lambda OKAY\n"); else printf("Lambd BAD\n");
      freemem(A);
      freemem(W);
      freemem(a);
      freemem(b);
      return 1;
}

if( 0 ) {

   printf("Diophantine solver test\n");
   p = 9223372036854775783LL;  // = 2^63-25
   p = 4611686018427387847LL;  // = 2^62-57
   printf("p:=%lld;\n",p);

   da = 5;
   db = 5;
   dc = da+db-1;
   a = array(da+1); A = array(da+1); 
   b = array(db+1); B = array(db+1);
   c = array(dc+1);
   g = array(da+1); s = array(db); t = array(da);
   W = array(da+2*db);
   sigma = array(db); tau = array(da);

   randvecp(da+1,a,p);  //printf("a:="); polprint64s(a,da); printf(";\n");
   randvecp(db+1,b,p);  //printf("b:="); polprint64s(b,db); printf(";\n");
   randvecp(dc+1,c,p);  //printf("c:="); polprint64s(c,dc); printf(";\n");
 
   P = recip1( p );

   ST = realtime();
   polcopy64s(a,da,A);
   polcopy64s(b,db,B);
   //polgcdext64s( A, B, da, db, g, s, t, &dg, &ds, &dt, W, p );
   polgcdext64s( A, B, da, db, g, s, 0, &dg, &ds, &dt, W, p );
   ET = realtime();
   printf("EEA time = %10.3f\n",ET-ST);

   //printf("g:="); polprint64s(g,dg); printf(";\n");
   //printf("s:="); polprint64s(s,ds); printf(";\n");
   //printf("t:="); polprint64s(t,dt); printf(";\n");
   ST = ET;
   good = poldiosol64s( a, b, c, s, da, db, dc, ds, sigma, tau, &dsigma, &dtau, W, p );
   TT = realtime();
   printf("Dio time = %10.3f\n",TT-ST);

   //printf("sigma:="); polprint64s(sigma,dsigma); printf(";\n");
   //printf("tau:="); polprint64s(tau,dtau); printf(";\n");

   freemem(a); freemem(A);
   freemem(b); freemem(B);
   freemem(c);
   freemem(g); freemem(s); freemem(t);
   freemem(W);
}


   
if( 0 ) {  // MTSHL n=4
       

   LONG *A, *X, *F, *Y, *G, *Z, *Aj, *Xj, **BETA, *Beta, *B, *U, *C;
     int dFx, dFy, dFz, dFw;
     int dGx, dGy, dGz, dGw;
     int dAx, dAy, dAz, dAw;
     int n, nf, ng, na, D[4], ns, nfalp, ngalp;
     LONG p, beta, alpha, *falpha, *galpha;
     LONG *Falp, *Yalp, *Galp, *Zalp;
     recint P;
     int mtshl;

      dFx = 4; dFy = 4; dFz = 4; dFw = 6; D[0]=dFx; D[1]=dFy; D[2]=dFz; D[3]=dFw;
      dGx = 4; dGy = 4; dGz = 4; dGw = 6; D[0]=dGx; D[1]=dGy; D[2]=dGz; D[3]=dGw;
      dFx = 4; dFy = 4; dFz = 4; dFw = 4; D[0]=dFx; D[1]=dFy; D[2]=dFz; D[3]=dFw;
      dGx = 4; dGy = 4; dGz = 4; dGw = 4; D[0]=dGx; D[1]=dGy; D[2]=dGz; D[3]=dGw;
      n = 4;
      p = 101; 
      printf("p := %lld;\n",p);
      P = recip1(p);
      nf = 6;
      F = array(nf); Y = array(nf); 
      randpoly64s(n, D, nf, F, Y, p); 
      //printsupp( Y, nf, n );
      nf = pol2simp64s( F, Y, nf, p );
      //printsupp( Y, nf, n );

      printf("F := "); polnsparseprint64s(F, Y, nf, n); printf(";\n");

      ng = 3;
      G = array(ng); Z = array(ng);
      randpoly64s(n, D, ng, G, Z, p); 
      ng = pol2simp64s( G, Z, ng, p );
      printf("G := "); polnsparseprint64s(G, Z, ng, n); printf(";\n");

      A = array(nf*ng);
      X = array(nf*ng);
      na = polsparsemul64s(F, Y, nf, G, Z, ng, p, A, X);
      printf("polsparsemul: #F=%d #G=%d #A=%d\n",nf,ng,na);
      printf("A := "); polnsparseprint64s(A, X, na, n); printf(";\n");


printf("MTSHL 4 variables \n");

      //printsupp( X, na, n );
      
      polnsparsevecdeg64s(X,na,n,D);
      dAx = D[3]; dAy = D[2]; dAz = D[1]; dAw = D[0];
      printf("deg(A):  x1=%d  x2=%d  x3=%d  x4=%d\n",dAx,dAy,dAz,dAw); 

      Aj = array(na);
      Xj = array(na);
      beta = 3; // rand64s(p);
      Beta = array(dAz+1);
      evalbeta(Beta,dAz,beta,P);
      printf("Beta = "); vecprint64s(Beta,dAz+1); printf(";\n");
      BETA = arrayofptrs(1);
      BETA[0] = Beta;
      evalinit( X, na, n, 3, 3, BETA, Aj, Xj, p );
      printf("Aj := "); polnsparseprint64s(A, Xj, na, n-1); printf(";\n");

      C = array(na);
      polcopy64s( A, na, C );
      B = array(na);
      U = array(na);
      ns = evalnext( C, Xj, Aj, na, B, U, p );
      printf("B := "); polnsparseprint64s(B, U, ns, 3); printf(";\n");

if( 0 )
      // test Horner expand
      { LONG **FF, **YY; int *NN; LONG *AA, *XX, *T1, *T2, *C, *U;
        LONG alpha;
        int na,nc;

      nc = 4;
      C = array(nc); U = array(nc); 
      randpoly64s(n, D, nc, C, U, p); 
        
        NN = arrayints(3); NN[0]=nf; NN[1]=ng; NN[2]=nc;
        FF = arrayofptrs(3); FF[0]=F; FF[1]=G; FF[2]=C;
        YY = arrayofptrs(3); YY[0]=Y; YY[1]=Z; YY[2]=U;
        AA = array(100); XX = array(100);
        T1 = array(100); T2 = array(100);
        alpha = 3;
        printf("n=%d\n",n);
        printf("F := "); polnsparseprint64s(FF[0],YY[0],NN[0],n); printf(";\n");
        printf("G := "); polnsparseprint64s(FF[1],YY[1],NN[1],n); printf(";\n");
        printf("C := "); polnsparseprint64s( C, U, nc, n ); printf(";\n");
        na = HornerExpand( 2, alpha, n, NN, FF, YY, AA, XX, T1, T2, p );
        printf("A := "); polnsparseprint64s(AA,XX,na,n+1); printf(";\n");
     }


      // Compute f(x1,x2,...xn-1,alphan), g(x1,x2,...,xn-1,alphan)
      alpha = 3;
      printf("alpha := %lld\n", alpha);
      falpha = array(dFw+1);
      evalbeta( falpha, dFw, alpha, P );
      Falp = array(nf);
      Yalp = array(nf);
      nfalp = polneval1(F, Y, nf, falpha, Falp, Yalp, n, p );
printf(" Falp := "); polnsparseprint64s(Falp, Yalp, nfalp, n-1); printf(";\n");
printf(" eval(F,x%d=alpha) mod p; \n",n);
      galpha = array(dGw+1);
      evalbeta( galpha, dGw, alpha, P );
      Galp = array(ng);
      Zalp = array(ng);
      ngalp = polneval1(G, Z, ng, galpha, Galp, Zalp, n, p );
printf(" Galp := "); polnsparseprint64s(Galp, Zalp, ngalp, n-1); printf(";\n");
printf(" eval(G,x%d=alpha) mod p; \n",n);

      printf("\n MTSHL:  j=%d  1 core \n\n", n);
      mtshl = MTSHL(1,A,X,na,Falp,Yalp,nfalp,Galp,Zalp,ngalp,n,alpha,p);
      printf("\n MTSHL done \n");

   }

if( 1 )
   { LONG *A, *X, *F, *Y, *G, *Z;
     int d, mtshl;
     int n, nf, ng, na, D[16], nfalp, ngalp;
     LONG p, alpha, *falpha, *galpha;
     LONG *Falp, *Yalp, *Galp, *Zalp;
     recint P; 

      n = 6; d = 5;
      n = 12; d = 5;
      if( ARGS[2]!=0 ) n = ARGS[2];
      if( ARGS[3]!=0 ) d = ARGS[3];

if( n<2 ) { printf("The number of variables must be at least 2.\n"); exit(1); }

      for( i=0; i<n; i++ ) D[i] = d;

      p = 101;
      p = 997; 
      p = 4611686018427387905LL;
      p = 5; p = (p<<55)+1;  // not working
      p = 7*(1<<26)+1; 
      p = 2147352577; // = 2^17 3 42 127 + 1
      p = 2147483647; // = 2^31 - 1
      if( ARGS[6]!=0 ) p = ARGS[6];

      P = recip1(p);
      nf = 5;
      nf = 100;
      nf = 1000;
      if( ARGS[4]!=0 ) nf = ARGS[4];
      ng = 5;
      ng = 100;
      ng = 1000;
      if( ARGS[4]!=0 ) ng = ARGS[4];
      if( ARGS[5]!=0 ) ng = ARGS[5];

printf("n=%d  d=%d  #f=%d  #g=%d\n",n,d,nf,ng);
printf("p := %lld;\n",p);

      F = array(nf); Y = array(nf); 
      randpoly64s(n, D, nf, F, Y, p); 
      nf = pol2simp64s( F, Y, nf, p );

      //printf("F := "); polnsparseprint64s(F, Y, nf, n); printf(";\n");
      //evalhens( F, Y, nf, n );
      //printf("F := "); polnsparseprint64s(F, Y, nf, n); printf(";\n");

      G = array(ng); Z = array(ng);
      randpoly64s(n, D, ng, G, Z, p); 
      ng = pol2simp64s( G, Z, ng, p );
      // printf("G := "); polnsparseprint64s(G, Z, ng, n); printf(";\n");

      A = array(nf*ng);
      X = array(nf*ng);

      ST = realtime();
      na = polsparsemul64s(F, Y, nf, G, Z, ng, p, A, X);
      ET = realtime();

      printf("\nTIME fxg = %9.3f s\n\n",ET-ST);

      //printf("A := "); polnsparseprint64s(A, X, na, n); printf(";\n");
      
      polnsparsevecdeg64s(X,na,n,D);
      printf("deg(A):");
      for( i=0; i<n; i++ ) printf("  x[%d]=%d", i, D[i]); printf("\n");

      // Compute f(x1,x2,...xn-1,alphan), g(x1,x2,...,xn-1,alphan)
      alpha = 3; //printf("alpha := %lld\n", alpha);
      falpha = array(d+1);
      evalbeta( falpha, d, alpha, P );
      Falp = array(nf);
      Yalp = array(nf);
      nfalp = polneval1(F, Y, nf, falpha, Falp, Yalp, n, p );
//printf(" Falp := "); polnsparseprint64s(Falp, Yalp, nfalp, n-1); printf(";\n");
//printf(" eval(F,x[%d]=alpha) mod p; \n",n);

      galpha = array(d+1);
      evalbeta( galpha, d, alpha, P );
      Galp = array(ng);
      Zalp = array(ng);
      ngalp = polneval1(G, Z, ng, galpha, Galp, Zalp, n, p );
//printf(" Galp := "); polnsparseprint64s(Galp, Zalp, ngalp, n-1); printf(";\n");
//printf(" eval(G,x[%d]=alpha) mod p; \n",n);

      printf("\n");
      mtshl = MTSHL(NumCores,A,X,na,Falp,Yalp,nfalp,Galp,Zalp,ngalp,n,alpha,p);
      printf("n=%d  d=%d  t=%d  NumCores = %d;\n", n,d,nf,NumCores);
      printf("#f=%d  #g=%d  #a=%d\n",nf,ng,na);

   }

   printf("exiting\n");

   return 0;

}
