// Copyright  Michael Monagan  May 2020
// This file contains HenselLift1 a bivariate Hensel lift in Zp[x1,x2] and
// HenselLift2 for Zp[x1,x2,x3] which uses dense interpolation on for x2
// This file was created by making a copy of hensel11.cilk stripping out the
// Cilk directives spawn, sync, cilk using  s/spawn//  s/sync//  s/cilk / /


#include <stdio.h>

#define DEBUG 0

#define LONG long long int
#define add64s add64hensel
#define sub64s sub64hensel
#define neg64s neg64hensel
#define mul64s mul64hensel
#define max32s max32hensel
#define min32s min32hensel

int max32s(int a, int b) { if(a>b) return a; else return b; }
int min32s(int a, int b) { if(a<b) return a; else return b; }
LONG add64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
LONG sub64s(LONG a, LONG b, LONG p) { LONG t; t = a-b; t += (t>>63) & p; return t; }
LONG neg64s(LONG a, LONG p) { LONG t; t = -a; t += (t>>63) & p; return t; }
LONG mul64s(LONG a, LONG b, LONG p) {
        LONG q, r;
        __asm__ __volatile__(           \
        "       mulq    %%rdx           \n\t" \
        "       divq    %4              \n\t" \
        : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
        return r;
}


LONG modinv64s( LONG c, LONG p );
LONG powmod64s( LONG a, LONG n, LONG p );

#define UINT64 unsigned long long
typedef struct {
        UINT64 s;       /* shift */
        UINT64 v;       /* reciprocal of d */
        UINT64 d0;      /* divisor shifted up */
        UINT64 d1;
} recint;

recint recip1(UINT64  p);
UINT64 mulrec64(UINT64  a, UINT64  b, recint  v);
void sort2vec64s(LONG *D, LONG *C, int n );


double realtime();

int * arrayints( int n );
LONG * array( int n );
int ** arrayptrsints( int n );
LONG ** arrayofptrs( int n );
LONG ** array2( int n, int m );
int ** array2ints( int n, int m );
void freemem(void *p);

LONG rand64s(LONG p);
void polprint64s( LONG *A, int d );
void vecprint64s( LONG *A, int n );
void vecpow64s( int n, LONG *A, LONG alpha, LONG p );
void polnsparseprint64s( LONG *C, LONG *X, int t, int n );
int polmul64s( LONG * A, LONG * B, LONG * C, int da, int db, LONG p);
int poldiv64s( LONG * A, LONG * B, int da, int db, LONG p );
void polcopy64s( LONG *A, int d, LONG *B );
void veccopy64s( LONG* A, int n, LONG *B );
LONG rand64s(LONG p);
int poladd64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG p);
int polsub64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG p);
int poladdsca64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG alpha, LONG p, recint P);
int polsubsca64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG alpha, LONG p, recint P);
void polscamul64s( LONG x, LONG *A, int d, LONG p );
int polnsparsedeg64s( LONG *X, int t, int n, int xi );
void polnsparsevecdeg64s( LONG *X, int t, int n, int *D );
LONG poleval64s(LONG *a, int d, LONG x, LONG p);
void VECSCAMUL64s( LONG x, LONG *A, int n, LONG p );
void VECSMULREC64s( LONG *A, int n, LONG x, recint P );
LONG polnsparseeval64s(LONG *C, LONG *X, int t, LONG *A, int n, LONG p );
void evalbeta( LONG *A, int d, LONG beta, recint P );
LONG getomega64s( LONG p, LONG n );
void FFT64s1( int n, LONG *a, LONG *W, LONG p, recint P );
void FFT64s2( int n, LONG *a, LONG *W, LONG p, recint P );
void MakeW64( int n, LONG w, LONG *W, LONG p );





void polgcdext64s( LONG *A, LONG *B, int da, int db,
                   LONG *G, LONG *S, LONG *T, int *dG, int *dS, int *dT, 
                   LONG *W, LONG p );

int poldiosol64s( LONG *A, LONG *B, LONG *C, LONG *S, int da, int db, int dc, int ds, // input
                  LONG *sigma, LONG *tau, int *dsigma, int *dtau, // output
                  LONG *W, /* temporary space */ LONG p )
{ 
   // solve sigma A + tau B = C
   // dc < da+db, ds < db, dsigma < db, dtau < da
   // compute W = (c mod b) s mod b in W with (c mod b) first computed in W
   // compute tau = (c-sigma a)/b in W
   // |W| > dc and |W| > 2db-1 and |W| > db-1 + da = max( da+db, 2db )
   int d;
   if( dc >= da+db || ds >= db ) { 
       printf("poldiosol64s: degrees wrong: da=%d db=%d dc=%d ds=%d\n", da,db,dc,ds);
       return 1;
   }
//printf("poldio: A= "); polprint64s(A,da); printf(";\n");
//printf("poldio: B= "); polprint64s(B,db); printf(";\n");
//printf("poldio: C= "); polprint64s(C,dc); printf(";\n");
//printf("poldio: S= "); polprint64s(S,ds); printf(";\n");
   polcopy64s( C, dc, W );
   d = poldiv64s( W, B, dc, db, p );        // W = C mod B
//printf("POLDIOSOL: mult = %d x %d  dc=%d  ",d,ds,dc);
   d = polmul64s( W, S, W, d, ds, p );
   dsigma[0] = poldiv64s( W, B, d, db, p );  // W = C S mod B
   polcopy64s( W, dsigma[0], sigma );

   d = polmul64s( A, W, W, da, dsigma[0], p );
   d = polsub64s( C, W, W, dc, d, p );
   if( poldiv64s( W, B, d, db, p ) != -1 ) { printf("poldiosol64s: division failed\n"); return 1; }

if( d-db < 0 ) dtau[0] = -1;
else {
   polcopy64s( W+db, d-db, tau );
   dtau[0] = d-db;
}
//printf("POLDIOSOL: dsigma=%d  dtau=%d\n ",dsigma[0],dtau[0]);

if( DEBUG )  { LONG *T;
    // check sigma a + tau b = c
    T = array(1000);
    polmul64s( sigma, A, W, dsigma[0], da, p );
    polmul64s( tau, B, T, dtau[0], db, p );
    d = poladd64s( W, T, W, dsigma[0]+da, dtau[0]+db, p );
    d = polsub64s( C, W, T, dc, d, p );
    if( d!=-1 ) { printf("poldio is bad\n"); return 1; } // else printf("poldio is good\n");
    freemem(T);
   }

   return 0;

}

                  
void randvecp( int n, LONG * A, LONG p ) {
   int i;
   for( i=0; i<n; i++ ) A[i] = rand64s(p);
   return;
}


LONG ** polmake2rec64s( LONG *A, LONG *X, int t ) {
   // [ dx, [ dy, c0, c1, ..., cdy ] x^0, ...  recursive dense
   int i,j,dx,dy,D[2],numbits;
   LONG ** B, *C, mask;
   polnsparsevecdeg64s( X, t, 2, D );
   dy = D[0];
   dx = D[1];
   B = arrayofptrs( dx+2 );
   mask = dx;
   B[0] = (LONG *) (mask);
   B++; // shift forward
   for( i=0; i<=dx; i++ ) B[i] = NULL;
   numbits = 63/2;
   mask = (1LL<<numbits)-1;
   j=t-1;
   while( j>=0 ) {
         dy = X[j] & mask;
         dx = X[j] >> numbits;
         C = array(dy+2);
         B[dx] = C;
         C[0] = dy;
         C ++;
         for(i=0; i<dy; i++) C[i] = 0;
         C[dy] = A[j];
         j --;
         while( j>=0 && (X[j] >> numbits) == dx ) {
            dy = X[j] & mask;
            C[dy] = A[j];
            j--;
        }
    }
    B--; // shift back!
    return B;
}



void polrec2conv64s( LONG *A, LONG *X, int t, LONG ***B, int **DegB ) {
   int i,j,dx,dy,x,y,D[2],numbits;
   LONG *C, mask;
   polnsparsevecdeg64s( X, t, 2, D );
   dy = D[0];
   dx = D[1];
   //printf("dx=%d dy=%d\n",dx,dy);

   DegB[0] = arrayints( dx+2 );
   DegB[0][0] = dx;
   for( i=0; i<=dx; i++ ) DegB[0][i+1] = dy;

   B[0] = array2( dx+1, dy+1 );
   for( i=0; i<=dx; i++ )
      for( j=0; j<=dy; j++ ) 
         B[0][i][j] = 0;

   numbits = 63/2;
   mask = (1LL<<numbits)-1;
   j=t-1;
   for( j=0; j<t; j++ ) {
         y = X[j] & mask;
         x = X[j] >> numbits;
         B[0][x][y] = A[j];
   }
   for( i=0; i<=dx; i++ ) {
      C = B[0][i];
      for( j=dy; j>=0 && C[j]==0; j-- );
      DegB[0][i+1] = j;
   }
   return;
}


int polmul2cof64s( LONG **A, LONG **B, int k, LONG *C, LONG *T, int p ) {
// coeff( A(x,y) B(x,y), x, k )  T and C must be large enough for the result
   int i,j,d,dC;
   LONG dA,dB,*a,*b;
   dA = (LONG) A[0]; A++;
   dB = (LONG) B[0]; B++;
   dC = -1;
   for( i=0,j=k; i<=dA && j>=0; i++,j-- ) {
      if( j>dB ) continue;
      a = A[i]; if( a==NULL ) continue;
      b = B[j]; if( b==NULL ) continue;
      d = polmul64s(a+1,b+1,T,a[0],b[0],p);
      dC = poladd64s(T,C,C,d,dC,p);
   }
   return dC;
}


int polrec2mulcof64s( LONG **A, int *DA, LONG **B, int *DB, int k, LONG *C, LONG *T, int p ) {
// coeff( A(x,y) B(x,y), x, k )  T and C must be large enough for the result
   int i,j,dA,dB,dC,d,da,db;
   dA = DA[0]; DA++; //printf("deg(a,x)=%d\n",dA);
   dB = DB[0]; DB++; //printf("deg(b,x)=%d\n",dB);
   dC = -1;
   for( i=0,j=k; i<=dA && j>=0; i++,j-- ) {
      if( j>dB ) continue;
      da = DA[i]; if( da==-1 ) continue; //printf("deg( cof(A,x^%d), y ) = %d\n",i,da);
      db = DB[j]; if( db==-1 ) continue; //printf("deg( cof(B,x^%d), y ) = %d\n",j,db);
      d = polmul64s(A[i],B[j],T,da,db,p);
      dC = poladd64s(T,C,C,d,dC,p);
   }
   return dC;
}


void taymulcof64s( LONG **A, LONG **B, int k, int d, LONG *C, recint P, int p ) {
   int i,j,n;
/*
   for( n=0; n<=d; n++ ) { // C[n] = coeff( F(x=n,y) G(x=n,y), y^k )
      LONG s; s = 0;
      for( i=1,j=k-1; i<k; i++,j-- )
          s = add64s(s,mulrec64(A[i][n],B[j][n],P),p);
      C[n] = s;
   }
*/
   // To make memory access sequential I've transposed the above code
   for( n=0; n<=d; n++ ) C[n] = 0;
   for( i=1,j=k-1; i<k; i++,j-- )
      for( n=0; n<=d; n++ )
          C[n] = add64s(C[n],mulrec64(A[i][n],B[j][n],P),p);
          //C[n] = add64s(C[n],mul64s(A[i][n],B[j][n],p),p);
   return;
}


void taymulcof264s( LONG **A, LONG **B, int N, int k, LONG *C, recint P, int p ) {
   int i,j,n;
   LONG s, *a, *b;
   for( n=0; n<N; n++ ) { // C[n] = coeff( F(x=omega^n,y) G(x=omega^n,y), y^k )
      a = A[n];
      b = B[n];
      s = 0;
      for( i=1,j=k-1; i<k; i++,j-- ) s = add64s(s,mulrec64(a[i],b[j],P),p);
      C[n] = s;
   }
   return;
}


        /* z += a*b */
        #define zfma(z,a,b) do {        \
        unsigned long u,v;              \
        __asm__(                        \
        "       mulq    %%rdx           \n\t" \
        "       addq    %%rax, %0       \n\t" \
        "       adcq    %%rdx, %1       \n\t" \
                : "=&r"(z[0]), "=&r"(z[1]), "=a"(u), "=d"(v) : "0"(z[0]), "1"(z[1]), "a"(a), "d"(b));\
        } while (0)


        /* z = z % p safe */
        #define zmod(z,p) __asm__(\
        "       divq    %4      \n\t" \
        "       xorq    %0, %0  \n\t" \
                : "=a"(z[1]), "=d"(z[0]) : "a"(z[0]), "d"(z[1] < p ? z[1] : z[1] % p), "r"(p))


LONG zacc( LONG *A, LONG *B, int n, LONG p );
void taymulcof364s( LONG **A, LONG **B, int N, int k, LONG *C, recint P, int p ) {
   int n;
   for( n=0; n<N; n++ ) { // C[n] = coeff( F(x=omega^n,y) G(x=omega^n,y), y^k )
      C[n] = zacc( A[n], B[n], k, p );
   }
   return;
}


void polprint2rec64s( LONG **A ) {
   int dx,dy,i,j,first;
   LONG *C,temp;
   first = 1;
   temp = (LONG) A[0];
   dx = temp;
   A++;
   for( i=0; i<=dx; i++ ) 
      if( A[i]!=NULL ) { 
         C = A[i]; dy = C[0]; C++;
         for( j=0; j<=dy; j++ ) {
            if( C[j]==0 ) continue;
            if( !first ) printf("+"); else first = 0;
            printf("%lld",C[j]);
            if( i!=0 ) printf("*x^%d",i);
            if( j!=0 ) printf("*y^%d",j);
         }
      }
   return;
}


void polrec2print64s( int *D, LONG **A ) {
// D[0] = dx, D[1]...D[dx+1] = degrees of coefficients
// A[i] = array of coefficients in y of x^i
   int dx,dy,i,j,first;
   LONG *C;
   first = 1;
   dx = D[0];
   if( dx==-1 ) { printf("0"); return; }
   first = 1;
   for( i=0; i<=dx; i++ ) {
      dy = D[i+1]; C = A[i];
      if( dy==-1 ) continue;
      if( C==NULL ) { printf("\nerror in pol2recprint64s: NULL pointer\n"); return; }
      for( j=0; j<=dy; j++ ) {
          if( C[j]==0 ) continue;
          if( !first ) printf("+"); else first = 0;
          printf("%lld",C[j]);
          if( i!=0 ) printf("*x^%d",i);
          if( j!=0 ) printf("*y^%d",j);
      }
   }
   return;
}




int polnewton2eval64s( int dy, int *D, LONG **U, LONG *M, LONG beta, LONG *B, LONG p )
// U = U[0](x)+U[1](x)(y-M[0])+U[2](x)(y-M[0])(y-M[1])+...   deg(T,x) >= deg(U[i],x)
// B(x) = eval(U(x,y),y=beta) mod p.
// return deg(B,x)
{
   int dx,db,i;
   LONG m;
   recint P;
   if( dy==-1 ) return -1;
   db = D[0]; // = deg(A[0],x)
   polcopy64s(U[0],db,B);
   P = recip1(p);
   for( m=1,i=1; i<=dy; i++ ) {
      m = mulrec64(m,sub64s(beta,M[i-1],p),P);
      if( m==0 ) printf("ERROR: ZERO in polnewton2eval64s\n");
      dx = D[i];
      if( dx>=0 ) db = poladdsca64s(B, U[i], B, db, dx, m, p, P); // B = B + m U[i]
   }
   return db;
}


int cilkpolnewton2eval64s( int dy, int *D, LONG **U, LONG *M, LONG beta, LONG *B, LONG *Fxy, LONG Mbeta, LONG dfj, LONG p )
{ int dx;
             dx = polnewton2eval64s( dy, D, U, M, beta, B, p );
             dx = polsub64s( Fxy, B, Fxy, dfj, dx, p );
             polscamul64s( Mbeta, Fxy, dx, p );
             return dx;
}



void polrec2tayprint64s( int *D, LONG **A, LONG alpha ) {
// D[0] = dx, D[1]...D[dx+1] = degrees of coefficients
// A[i] = array of coefficients in y of x^i
   int dx,dy,i,first;
   LONG *C;
   first = 1;
   dx = D[0];
   if( dx==-1 ) { printf("0"); return; }
   first = 1;
   for( i=0; i<=dx; i++ ) {
      dy = D[i+1]; C = A[i];
      if( dy==-1 ) continue;
      if( C==NULL ) { printf("\nerror in pol2recprint64s: NULL pointer\n"); return; }
      if( first ) first = 0; else printf("+");
      printf("("); polprint64s(C,dy); printf(")");
if( alpha==0 ) { 
      if( i==1 ) printf("*z");
      if( i>=2 ) printf("*z^%d",i);
}
else {
      if( i==1 ) printf("*(z-%lld)",alpha);
      if( i>=2 ) printf("*(z-%lld)^%d",alpha,i);
}
   }
   return;
}


void pol2zprint( int *D, LONG **f, LONG alpha) {
    // debugging utility to print a0(x)+a1(x)(z-alpha)+a2(x)(z-alpha^2)
    LONG i,d;
    d = D[0];
    for( i=0; i<=d; i++ )  {
        if( i>0 ) printf("+");
        printf("("); polprint64s(f[i],D[i+1]); printf(")");
        if( i>0 ) 
            if( alpha==0 ) printf("*z^%lld",i);
            else printf("*(z-%lld)^%lld",alpha,i);
    }
    return;
}

int poladdsca64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG alpha, LONG p, recint P);
int polsubsca64s(LONG *a, LONG *b, LONG *c, int da, int db, LONG alpha, LONG p, recint P);

void hornerexpand64s( int *D, LONG **A, LONG alpha, LONG p ) {
// D[0] = dz
// A[i] = array of coefficients in Zp[x] for 0 <= i <= dz.
// D[i+1] = degree(A[i],x) for 0 <= i <= dz
// expand( sum( A[i]*(z-alpha)^i, i=0..dz ) ) inplace
    LONG i, j, beta, dz;
    recint P;
    P = recip1(p);
    beta = neg64s(alpha,p);
    dz = D[0];
    if( dz<0 ) printf("ERROR in hornerexpand: dz<0\n");
    //printf("p := %lld;\n",p);
    //printf("f := "); pol2zprint(D,A,alpha); printf(";\n");
    for( i=dz-1; i>=0; i-- ) 
        for( j=i; j<dz; j++ )
            D[j+1] = poladdsca64s(A[j],A[j+1],A[j],D[j+1],D[j+2],beta,p,P);
    //printf("g := "); pol2zprint(D,A,0); printf(";\n");
    return;
}


void pol2tayprint64s( int dy, int *D, LONG **A, LONG alpha ) {
// D[0] = dx, D[1]...D[dx+1] = degrees of coefficients
// A[i] = array of coefficients in y of x^i
   int dx,i,first;
   LONG *C;
   if( dy==-1 ) { printf("0"); return; }
   first = 1;
   for( i=0; i<=dy; i++ ) {
      dx = D[i];
      if( dx==-1 ) continue;
      C = A[i];
      if( C==NULL ) { printf("\nerror in pol2recprint64s: NULL pointer\n"); return; }
      if( first ) first = 0; else printf("+");
      printf("("); polprint64s(C,dx); printf(")");
      if( i==1 ) printf("*(y-%lld)",alpha);
      if( i>=2 ) printf("*(y-%lld)^%d",alpha,i);
   }
   return;
}



int poldiff64s( LONG *A, LONG *X, int t, int n, int j, LONG p ) {
// diff(a,x[j]) a is stored as pair (A,X) with t terms in n variables
   int i, k, numbits, shift; LONG mask, x, one; recint P;
   P = recip1( p );
   numbits = 63/n;
   shift = numbits*(n-j);
   mask = (1LL<<numbits)-1;
   one = 1LL << shift;
   k = 0;
   for( i=0; i<t; i++ ) {
       x = (X[i] >> shift) & mask;
       if( x>0 ) { A[k] = mulrec64(x,A[i],P); X[k] = X[i] - one; k++; }
   }
   return k; // number of terms in derivative
}


void evalpowers64s( LONG alpha, int d, LONG *W, LONG p ) 
{ return vecpow64s( d+1, W, alpha, p ); }


LONG evalhorner( LONG *f, LONG d, LONG x, recint P, LONG p ) {
    // f(x) mod p
    int i; LONG c;
    if( d<0 ) return 0;
    c = f[d];
    for( i=d-1; i>=0; i-- ) c = add64s(f[i],mulrec64(c,x,P),p);
    return c;
}


void evalf0tod( LONG *f, int df, int d, LONG *Y, LONG p ) {
// Y[i] = f(i), i=0..d  where df = deg(f) 
    LONG i; recint P;
    P = recip1(p);
    for( i=0; i<=d; i++ ) Y[i] = evalhorner( f, df, i, P, p );
    return;
}


int fftgetn( int d ) {
    int n;
    for( n=1; n<=d; n=2*n );
    return n;
}


void ffteval( LONG *f, int df, LONG *W, int n, LONG *Y, LONG p, recint P ) {
    int i;
    if( df<-1 ) printf("ERROR in ffteval: df=%d\n",df);
    if( df<0 ) { for(i=0; i<n; i++ ) Y[i] = 0; return; }
    for( i=0; i<=df; i++ ) Y[i] = f[i];
    for( i=df+1; i<n; i++ ) Y[i] = 0;
    FFT64s1( n, Y, W, p, P );
    return;
}


int fftinterp( LONG *Y, int n, LONG *WINV, LONG *C, LONG p, recint P ) {
    int i,d; LONG ninv;
    FFT64s2( n, Y, WINV, p, P );
    for( d=n-1; d>=0 && Y[d]==0; d--);
    ninv = modinv64s((LONG) n, p);
    for( i=0; i<=d; i++ ) C[i] = mulrec64(ninv,Y[i],P);
    return d;
}
 

void newt2stand( LONG *V, int d, LONG *B, recint P, LONG p ) {
    int i,j; LONG alpha;
    B[0] = V[d];
    for( i=0,alpha=d-1; alpha>=0; alpha--,i++ ) {
        // B = (x-alpha) B mod p
        B[i+1] = B[i];
        for( j=i; j>0; j-- ) B[j] = sub64s(B[j-1],mulrec64(alpha,B[j],P),p);
        B[0] = sub64s(V[alpha],mulrec64(alpha,B[0],P),p);
        //printf(" B = "); vecprint64s(B,i+2); printf("\n");
    }
    return;
}

int newtoncofs( LONG *U, int d, LONG *minv, LONG *V, recint P, LONG p ) {
// Let V = V[0]+(x-0)V[1]+(x-1)(x-0)V[2]+(x-2)(x-1)(x-0)V[3]+...+(x-d-1)...(x-1)(x-0)V[d]
// Let V(i) = U[i] so that 
// V[i]=(U[i] - (V[0]+iV[1]+i(i-1)V[2]+...+i!V[i-1])) / i!  which in horner form is
// V[i]=(U[i] - (V[0]+i(V[1]+(i-1)(V[2]+...+3(V[i-2]+2(V[i-1]))))) / i!
    int i,j; LONG t,m;
    if( d<0 ) return d;
    V[0] = U[0];
    for( i=1; i<=d; i++ ) {
        t = V[i-1];
        for( m=2,j=i-2; j>=0; j--,m++ )
            t = add64s( mulrec64(m,t,P), V[j], p );
        t = mulrec64( sub64s(U[i],t,p), minv[i], P );
        V[i] = t;
    }
    while( d>=0 && V[d]==0 ) d--;
    return d;
}

int interp0tod64s( LONG *Y, int d, LONG *FACTINV, LONG *V, LONG *F, LONG p ) {
//  Given F(i)=Y[i] for 0<=i<=d interpolate F(x) from (i,Y[i]) in F and return deg(F)
    recint P;
//extern double NewtT;
//extern double ConvT;
//double T1,T2;
    if( d<0 ) return d;
    P = recip1(p);
    //F[0] = 1; // compute F[i] = 1/i! mod p = F[i-i]/i mod p
    //for( i=1; i<=d; i++ )
    //    F[i] = mulrec64( F[i-1], modinv64s((LONG) i,p), P );
    //d = newtoncofs( Y, d, F, V, P, p );
//T1 = realtime();
    d = newtoncofs( Y, d, FACTINV, V, P, p );
//T2 = realtime();
//NewtT += T2-T1;
    newt2stand( V, d, F, P, p );
//T1 = realtime();
//ConvT += T1-T2;
    return d;
}



int polbivareval64sold( LONG *A, LONG *X, int na, LONG *W, LONG *B, int dx, int dy, LONG p ) {
// eval(a(x,y),y=j) a is stored as pair (A,X) with t terms in 2 variables
// Output is a dense array of coefficients of degree dx
   int i,shift,x,y;
   LONG mask;
   recint P;
   P = recip1(p);
//printf("BIVAR  dx=%d  dy=%d  p=%lld", dx,dy,p); vecprint64s(W,dy); printf(";\n");
   for( i=0; i<=dx; i++ ) B[i] = 0;
   shift = 31;
   mask = (1LL<<31)-1;
   for( i=0; i<na; i++ ) {
       y = X[i] & mask;       if( DEBUG ) if ( y>dy ) printf(" **BIVAR degree dy BUG: y=%d  dy=%d**\n",y,dy);
       x = X[i] >> shift;     if( DEBUG ) if ( x>dx ) printf(" **BIVAR degree dx BUG: x=%d  dx=%d**\n",x,dx);
       B[x] = add64s(B[x],mulrec64(W[y],A[i],P),p);
   }
   while( dx>=0 && B[dx]==0 ) dx--;
   if( DEBUG ) { LONG Aval, Bval, BETA[2];
         BETA[0] = W[1];
         BETA[1] = rand64s(p);
         Aval = polnsparseeval64s(A, X, na, BETA, 2, p );
         Bval = poleval64s( B, dx, BETA[1], p );
         if( Aval!=Bval ) printf("BIVAR EVAL BUG\n");
   }
   return dx;
}



int polbivareval64s( LONG *A, LONG *X, int na, LONG *W, LONG *B, int dx, int dy, LONG p ) {
// eval(a(x,y),y=j) a is stored as pair (A,X) with t terms in 2 variables
// Output is a dense array of coefficients of degree dx
   int i,j,shift,x,y;
   LONG mask,z[2];
   recint P;
   P = recip1(p);
//printf("BIVAR  dx=%d  dy=%d  p=%lld", dx,dy,p); vecprint64s(W,dy); printf(";\n");
   for( i=0; i<=dx; i++ ) B[i] = 0;
   shift = 31;
   mask = (1LL<<31)-1;
   for( i=0; i<na; i=j ) {
       x = X[i] >> shift;
       //for( t=B[x]; i<j; i++ ) { y = X[i] & mask; t = add64s(t,mulrec64(W[y],A[i],P),p); }
       //B[x] = t;
       for( j=i+1; j<na && ((X[j] >> shift) == x); j++ );
       z[0] = B[x];
       z[1] = 0;
       while( i<j-1 ) {
            y = X[i] & mask; zfma(z,W[y],A[i]); if( z[1]>=p ) z[1] -= p; i++;
            // 20% faster to have if( z[1]>=p ) here in the middle than at the beginning or end of the loop
            y = X[i] & mask; zfma(z,W[y],A[i]); i++;
       }
       if( i==j-1 ) { y = X[i] & mask; zfma(z,W[y],A[i]); if( z[1]>=p ) z[1] -= p; i++; }
       zmod(z,p);
       B[x] = z[0];
   }
   while( dx>=0 && B[dx]==0 ) dx--;
   return dx;
}


int polbivarkeval64s( LONG *A, LONG *X, int na, LONG *ALPHA, LONG *B, int dx, int dy, int k, LONG p ) {
// eval(a(x,y),y=alpha) a is stored as pair (A,X) with t terms in 2 variables
// ALPHA[i] = alpha^i mod p
// Output in B is a dense array of coefficients of degree dx
   int i,j,n,shift,x,y;
   LONG mask,z[2];
   for( i=0; i<=dx; i++ ) B[i] = 0;
   shift = 31;
   mask = (1LL<<31)-1;
   for( i=0; i<na; i=n ) {
       x = X[i] >> shift;
       for( n=i+1; n<na && ((X[n] >> shift) == x); n++ );
       z[0] = B[x];
       z[1] = 0;
if( p<=mask ) // p<2^31 so we won't overflow a 128 bit register
       for( j=n-1; j>=i && (y = X[j] & mask) >= k; j-- ) { zfma(z,ALPHA[y-k],A[j]); }
else   for( j=n-1; j>=i && (y = X[j] & mask) >= k; j-- ) { zfma(z,ALPHA[y-k],A[j]); if( z[1]>=p ) z[1] -= p; }
       zmod(z,p);
       B[x] = z[0];
   }
   while( dx>=0 && B[dx]==0 ) dx--;
   return dx;
}


void polbivarsca64s( LONG *A, LONG *X, int na, LONG *ALPHA, recint P ) {
// a := a(x,alpha*y) mod p
   int i,y; LONG mask;
   mask = (1LL<<31)-1;
   for( i=0; i<na; i++ ) {
       y = X[i] & mask;
       A[i] = mulrec64(ALPHA[y],A[i],P);
   }
   return;
}


int polbivar1eval64s( LONG *A, LONG *X, int na, LONG *B, int dx, LONG p ) {
// eval(a(x,y),y=1) a is stored as pair (A,X) with t terms in 2 variables
// Output is a dense array of coefficients of degree dx
   int i,j,shift,x; LONG t;
   for( i=0; i<=dx; i++ ) B[i] = 0;
   shift = 31;
   for( i=0; i<na; i=j ) {
       x = X[i] >> shift;
       t = A[i];
       for( j=i+1; j<na && ((X[j] >> shift) == x); j++ ) t = add64s(t,A[j],p);
       B[x] = t;
   }
   while( dx>=0 && B[dx]==0 ) dx--;
   return dx;
}



int poltrivareval64s( LONG *A, LONG *X, int na, LONG *beta, LONG *B, LONG *Y, LONG p ) {
// eval(a,x[2]=beta) a is stored as pair (A,X) with t terms in n variables
// beta[i] = alpha^i for some alpha
// Output is a sparse polynomial in x,z
   int i,shift,nb;
   LONG mask,x,y,z,c;
   recint P;
   P = recip1( p );
   shift = 21;
   mask = (1LL<<21)-1;
   for( i=0; i<na; i++ ) {
       z = X[i] & mask;
       x = X[i] >> shift;
       y = x & mask;
       x = x >> shift;
       B[i] = mulrec64(A[i],beta[y],P); 
       Y[i] = (x<<31)+z;
   }
   sort2vec64s(Y, B, na ); // do we need this -- not if x1 > xj > x2
   nb = 0;
   for( i=0; i<na; ) { // add up adjacent terms
       x = Y[i];
       c = B[i];
       for( i++; i<na && Y[i]==x; i++ ) c = add64s(c,B[i],p);
       if( c!=0 ) { B[nb] = c; Y[nb] = x; nb++; }
   }
   if( DEBUG ) for( i=0; i<nb-1; i++ ) if( Y[i]>Y[i+1] ) printf("NOT SORTED\n");
   if( DEBUG ) { LONG Aval, Bval, BETA[3];
         BETA[0] = rand64s(p);
         BETA[1] = beta[1];
         BETA[2] = rand64s(p); 
         Aval = polnsparseeval64s(A, X, na, BETA, 3, p );
         BETA[1] = BETA[2];
         Bval = polnsparseeval64s(B, Y, nb, BETA, 2, p );
         if( Aval!=Bval ) printf("TRIVAL BUG\n");
   }
   return nb;
}


int poltrivar3eval64s( LONG *A, LONG *X, int na, LONG *beta, LONG *B, LONG *Y, LONG p ) {
// eval(a,x[3]=beta) a is stored as pair (A,X) with t terms in n variables
// beta[i] = alpha^i for some alpha
// Output is a sparse polynomial in x,z
   int i,shift,nb;
   LONG mask,x,y,z,c;
   recint P;
   P = recip1( p );
   shift = 21;
   mask = (1LL<<21)-1;
   for( i=0; i<na; i++ ) {
       z = X[i] & mask;
       x = X[i] >> shift;
       y = x & mask;
       x = x >> shift;
       B[i] = mulrec64(A[i],beta[z],P); 
       Y[i] = (x << 31) + y;
   }
   nb = 0;
   for( i=0; i<na; ) { // add up adjacent terms
       x = Y[i];
       c = B[i];
       for( i++; i<na && Y[i]==x; i++ ) c = add64s(c,B[i],p);
       if( c!=0 ) { B[nb] = c; Y[nb] = x; nb++; }
   }
   if( DEBUG ) { int sorted;
         sorted = 1;
         for( i=0; i<nb-1; i++ ) if( Y[i]>Y[i+1] ) sorted = 0;
         if( sorted ) printf("trivareval: SORTED\n"); else printf("trivareval: NOT SORTED\n");
   }
   if( DEBUG ) { LONG Aval, Bval, BETA[3];
         BETA[0] = beta[1];
         BETA[1] = rand64s(p);
         BETA[2] = rand64s(p); 
         Aval = polnsparseeval64s(A, X, na, BETA, 3, p );
         BETA[0] = BETA[1];
         BETA[1] = BETA[2];
         Bval = polnsparseeval64s(B, Y, nb, BETA, 2, p );
         if( Aval==Bval ) printf("TRIVAL 3 OK\n"); else printf("TRIVAL 3 BUG\n");
   }
   return nb;
}


int polrec2eval64s( LONG **A, LONG *B, LONG alpha, LONG p ) {
   LONG r, temp, *C;
   int dx,dy,i,j;
   recint P;
   P = recip1( p );
   temp = (LONG) A[0];
   dx = temp;
   A++;
   for( i=0; i<=dx; i++ ) {
      r = 0;
      if( A[i]!=NULL ) { 
          C = A[i]; dy = C[0]; C++;
          for( j=dy; j>=0; j-- ) r = add64s( C[j], mulrec64(alpha,r,P), p ); // Horner
      }
      B[i] = r;
   }
   while( dx>=0 && B[dx]==0 ) dx--;
   return dx;
}
 


int  hensel1check( LONG *A, LONG *X, int na, LONG **Fxy, int *dF, LONG **Gxy, int *dG, LONG alpha, LONG p )
// Test if A(x1,x2) = sum( F[i](x1) (y-alpha)^i, i=0..dFy ) sum( G[i](x) (y-alpha)^i, i=0..dGz ) at a random point mod p
{  
    int i,dFy,dGy; LONG Aval,Zval,Fval,Gval,beta[2],zero;
    dFy = dF[0]; dF ++;
    dGy = dG[0]; dG ++;
    beta[0] = rand64s(p);
    beta[1] = rand64s(p);
    Aval = polnsparseeval64s(A, X, na, beta, 2, p );
    Zval = sub64s(beta[0],alpha,p);
    Fval = poleval64s(Fxy[dFy],dF[dFy],beta[1],p);
    for( i=dFy-1; i>=0; i-- ) Fval = add64s( poleval64s(Fxy[i],dF[i],beta[1],p), mul64s(Fval,Zval,p), p );
    Gval = poleval64s(Gxy[dGy],dG[dGy],beta[1],p);
    for( i=dGy-1; i>=0; i-- ) Gval = add64s( poleval64s(Gxy[i],dG[i],beta[1],p), mul64s(Gval,Zval,p), p );
    zero = sub64s(Aval,mul64s(Fval,Gval,p),p);
//if( zero != 0 ) {
//printf("   In hensel1check  na=%d  dFy=%d  dGy=%d.\n",na,dFy,dGy);
//printf("   p := %lld;  alpha := %lld;\n",p,alpha);
//printf("   A1 := "); polnsparseprint64s( A, X, na, 2 ); printf(";\n");
//printf("   F1 := "); pol2tayprint64s( dFy, dF, Fxy, alpha ); printf(";\n");
//printf("   G1 := "); pol2tayprint64s( dGy, dG, Gxy, alpha ); printf(";\n");
//printf("   beta := "); vecprint64s(beta,2); printf(";\n");
//printf("   Aval = %lld\n", Aval );
//printf("   Fval = %lld\n", Fval );
//printf("   Gval = %lld\n", Gval );
//}
    return( zero==0 );
}



void factorialinv64s( LONG *F, int d, recint P, LONG p ) {
   LONG i;
   F[0] = 1;
   for( i=1; i<=d; i++ ) F[i] = mulrec64( F[i-1], modinv64s(i,p), P );
   return;
}




double DiffT, BevalT, InterpT, DioT, TaymulT, EvalT;


int HenselLift1( LONG *A, LONG *X, int na, LONG *f, int df, LONG *g, int dg, LONG alpha, LONG p,
                 LONG *SPACE, // temporary space
                 LONG ** Fxy, int *DegF, LONG ** Gxy, int *DegG,
                 LONG ** Fxyvals, LONG ** Gxyvals, LONG *Omega, LONG *OmegaInv, int N )

// a(x,y) stored as (A,X) with na terms
// f(x) of degree df
// g(x) of degree dg
// a(x,alpha) = f(x) g(x) mod p
// Fxy,DegF and Gxy,DegG is space for the result with y the main variable (lifted)
// SPACE is of size 8 (deg(a,x)+1) + (deg(a,y)+1)

{
   int i,k,dAx,dAy,dc,ds,dt,nAk,dG,dFy,dGy,dCx,dTx,flag,dsig,dtau;
   LONG *F1,*F2,*C,*G,*S,*T,*W,*V,*ALPHA,*Ak,*Xk,*sig,*tau,factinv,alphainv;
   recint P; 
   double T1,T2;
   

   dAx = polnsparsedeg64s( X, na, 2, 1 );
   dAy = polnsparsedeg64s( X, na, 2, 2 );
   if( df+dg != dAx ) {printf("error in HenselLift1: df+dg <> deg(A,x)"); return 0; }

   P = recip1(p);

if( DEBUG ) { LONG *BETA, *B, *C; int d, dc;
   printf("  Hensel1: dAx=%d  dAy=%d  df=%d  dg=%d\n",dAx,dAy,df,dg);
   if( 0 ) { printf("p := %lld; alpha := %lld; \n",p,alpha);
             printf("A := "); polnsparseprint64s(A, X, na, 2); printf(";\n");
             printf("f := "); polprint64s(f,df); printf(";\n");
             printf("g := "); polprint64s(g,dg); printf(";\n"); 
   }
   C = array(df+dg+1);
   dc = polmul64s( f, g, C, df, dg, p );
   B = array(dAx+1);
   BETA = array(dAy+1);
   evalbeta( BETA, dAy, alpha, P );
   d = polbivareval64s( A, X, na, BETA, B, dAx, dAy, p );
   d = polsub64s( C, B, C, dc, d, p );
   if( d!=-1 ) printf("Hensel 1: input BAD\n");
   freemem(B);
   freemem(C);
   freemem(BETA);
}

   { int d;
     d = dAx+1;
     F1 = SPACE; F2 = SPACE+d; S = SPACE+2*d; G = SPACE+3*d;
     W = SPACE+4*d; V = SPACE+5*d; T = SPACE+6*d; ALPHA = SPACE+8*d;
   }

   // solve s f + t g = 1
   polcopy64s(g,dg,F2);
   polcopy64s(f,df,F1);
   // solve S F2 + T F1 = G
   polgcdext64s( F2, F1, dg, df, G, S, NULL, &dG, &ds, &dt, W, p );
   if( dG>0 ) { printf("Hensel1: deg(G)>0\n"); return 0; }

   //printf("ds=%d\n",ds);
   //printf("S := "); polprint64s(S,ds); printf(";\n");
   //printf("T := "); polprint64s(T,dt); printf(";\n");


   // Initialize F = f and G = g
   DegF[0] = 0; sig = Fxy[0]; DegF[1] = df; polcopy64s(f,df,sig);
   DegG[0] = 0; tau = Gxy[0]; DegG[1] = dg; polcopy64s(g,dg,tau);


   C = F1; // reuse this space
   Ak = A;
   Xk = X;
   nAk = na;

   if( DEBUG > 0 ) {
      Ak = array(na); veccopy64s(A,na,Ak);
      Xk = array(na); veccopy64s(X,na,Xk);
   }

   evalpowers64s( alpha, dAy, ALPHA, p );  //  [ alpha^i mod p : 0<=i<=dAy ]
   polbivarsca64s( Ak, Xk, na, ALPHA, P );   // Ak(x,y) = Ak(x,alpha y)

   for( i=0; i<=dAy; i++ ) ALPHA[i] = 1;

   dFy = 0; dGy = 0; factinv = 1; alphainv = modinv64s(alpha,p);

   for( k=1; dFy+dGy < dAy; k++ ) {              //printf("  Hensel1: k=%d\n",k);

T2 = realtime();
      // nAk = poldiff64s( Ak, Xk, nAk, 2, 2, p );
      for( i=1; i<=dAy-k; i++ ) ALPHA[i] = add64s( ALPHA[i-1], ALPHA[i], p );
      //printf("ALPHA = "); vecprint64s(ALPHA,dAy-k+1); printf("\n");
T1 = realtime();
DiffT += T1-T2;

      // dCx = polbivareval64s( Ak, Xk, nAk, ALPHA, C, dAx, dAy, p );
      // dCx = polbivar1eval64s( Ak, Xk, nAk, C, dAx, p );
      dCx = polbivarkeval64s( Ak, Xk, nAk, ALPHA, C, dAx, dAy, k, p );

      if( k==1 ) factinv = alphainv; else factinv = mulrec64(alphainv,factinv,P);
      // factinv = (alpha^k)^(-1)
      VECSMULREC64s( C, dCx+1, factinv, P );

T2 = realtime();
BevalT += T2-T1;

  if( N ) {
      // ffteval( Fxy[k-1], DegF[k], Omega, N, Fxyvals[k-1], p, P );  // evaluate Fxy[k-1](x) at omega^i
      ffteval( Fxy[k-1], DegF[k], Omega, N, W, p, P );  // TRANS evaluate Fxy[k-1](x) at omega^i 
      { int j; for( j=0; j<N; j++ ) Fxyvals[j][k-1] = W[j]; }
      //ffteval( Gxy[k-1], DegG[k], Omega, N, Gxyvals[k-1], p, P );  // evaluate Gxy[k-1](x) at omega^i
      ffteval( Gxy[k-1], DegG[k], Omega, N, W, p, P );  // TRANS evaluate Gxy[k-1](x) at omega^i
      { int j; for( j=0; j<N; j++ ) Gxyvals[j][k-1] = W[j]; }
      //printf("k=%d  N=%d  DegF[k]=%d  degG[k]=%d\n",k,N,DegF[k],DegG[k]);
T1 = realtime();
EvalT += T1-T2;
      // taymulcof64s( Fxyvals, Gxyvals, k, N-1, W, P, p ); // W needs to be of size N not dAx+1
      taymulcof364s( Fxyvals, Gxyvals, N, k, W, P, p ); // TRANS W needs to be of size N not dAx+1
T2 = realtime();
TaymulT += T2-T1;
      dTx = fftinterp( W, N, OmegaInv, T, p, P );
T1 = realtime();
InterpT += T1-T2;
      //printf("dAx=%d  dTx=%d\n",dAx,dTx);
      if( dTx > dAx ) { printf( "Error in FFT eval/interp degree too big" ); }
  }
  else dTx = polrec2mulcof64s( Fxy, DegF, Gxy, DegG, k, T, W, p );

      if( DEBUG ) {
          dc = polrec2mulcof64s( Fxy, DegF, Gxy, DegG, k, G, W, p );
          if( dc!=dTx ) printf(" ERROR dc=%d dTx=%d\n",dc,dTx);
          for( i=0; i<=dc; i++ ) if( T[i]!=G[i] ) { printf(" ERROR( in eval/interp\n"); break; }
      }

      dc = polsub64s( C, T, C, dCx, dTx, p );

          //printf("C := "); polprint64s(C,dc); printf(";\n");
          //printf("            HENSEL 1: deg(A,x) = %d  deg(C,x) = %d\n", dAx, dc );

      sig = Fxy[k];
      tau = Gxy[k];
T1 = realtime();
      flag = poldiosol64s( g, f, C, S, dg, df, dc, ds, sig, tau, &dsig, &dtau, W, p );
T2 = realtime();
DioT += T2-T1;

      DegF[k+1] = dsig; if( dsig >= 0 ) { dFy = k; DegF[0] = k; }
      DegG[k+1] = dtau; if( dtau >= 0 ) { dGy = k; DegG[0] = k; }

          //printf("sigma := "); polprint64s(sig,dsig); printf(";\n");
          //printf("tau := "); polprint64s(tau,dtau); printf(";\n");
   }

          if( 0 ) { printf("  p := %lld;\n",p); }
          if( 0 ) { printf("  A := "); polnsparseprint64s(A, X, na, 2); printf(";\n"); }
          if( 0 ) { printf("  F := "); polrec2tayprint64s( DegF, Fxy, alpha ); printf(";\n"); }
          if( 0 ) { printf("  G := "); polrec2tayprint64s( DegG, Gxy, alpha ); printf(";\n"); }

hornerexpand64s( DegF, Fxy, alpha, p );
hornerexpand64s( DegG, Gxy, alpha, p );
//printf("  F := "); polrec2tayprint64s( DegF, Fxy, 0 ); printf(";\n");
//printf("  G := "); polrec2tayprint64s( DegG, Gxy, 0 ); printf(";\n");
//hornerexpand64s( DegF, Fxy, p-alpha, p );
//printf("  FF := "); polrec2tayprint64s( DegF, Fxy, alpha ); printf(";\n");



   if( DEBUG > 0 ) {
      //if( !hensel1check(A, X, na, Fxy, DegF, Gxy, DegG, alpha, p ) )
      if( !hensel1check(A, X, na, Fxy, DegF, Gxy, DegG, 0, p ) )  // ALPHA = 0
      {
            printf("Hensel1: FAIL\n");
            if( 0 ) { printf("  A := "); polnsparseprint64s(A, X, na, 2); printf(";\n"); }
            if( 0 ) { printf("  F := "); polrec2tayprint64s( DegF, Fxy, alpha ); printf(";\n"); }
            if( 0 ) { printf("  G := "); polrec2tayprint64s( DegG, Gxy, alpha ); printf(";\n"); }
      }
      else printf("Hensel1: OKAY\n");
      freemem(Ak); freemem(Xk);
   }

   return 1;
}


void polbivarimage64s( int dx, int dz, LONG ***F, int **D )
{
   int i; LONG *T, **G;
   D[0] = arrayints(dz+2);
   G = arrayofptrs(dz+1);
   T = array((dz+1)*(dx+1));
   for( i=0; i<=dz; i++ ) { G[i] = T; T = T+(dx+1); }
   F[0] = G;
   return;
}


void newton2standard64s( int dy, int *D, LONG **U, LONG *M, int p ) {
// D[i] = degree(U[i](x),x)
// Given U(x,y) = U0(x) + (y-M0) U1(x) + (y-M0)(y-M1) U2(x) + ... expand into
// expand to      V0(x) + y V1(x) + y^2 V2(x) + ...   inplace
   int i,k;
   recint P;
   P = recip1(p);
   for( k=dy-1; k>=0; k-- )
       for( i=k; i<dy; i++ ) {
//printf(" D[%d] = %d, D[%d] = %d : ", i, D[i], i+1, D[i+1] );
           D[i] = polsubsca64s( U[i], U[i+1], U[i], D[i], D[i+1], M[k], p, P );
           // U[i] = U[i] - M[k] U[i+1] in Zp[x]
//printf(" D[%d] = %d\n", i, D[i] );
       }
   return;
}


int cilknewton2standard64s( int dy, int *D, LONG **U, LONG *M, int p ) {
         newton2standard64s( dy, D, U, M, p );
         return 1;
}


int rec2bivar64s( int dy, int *D, LONG **F, LONG *A, LONG *X ) {
   int i,j,k,dx; LONG c,x,y;
// Given f(x,y) = sum( F[i][j] x^j y^i, j=0..D[i] ), i=0..dy )
// Make X = [ y^0x^0 | y^0x^1 | y^0x^2 | ... | y^1x^0 | y^1x^1 | ... | y^dyx^0 | y^dyx^1 | ... ]
// This choice with y first means the monomials are sorted in ascending order
                                                        //printf("Inrec2bivar:\n");
   for( k=0,i=0; i<=dy; i++ ) {
      y = i;
      dx = D[i];                                       //printf("dx=%d\n",dx); if( F[i]==NULL ) printf("F[i]==NULL\n",i);
      for( j=0; j<=dx; j++ ) {
         c = F[i][j];
         if( c!=0 ) { A[k] = c; x = j; X[k] = (y<<31)|x; k++; }
      }
   }
   if( DEBUG ) for( i=1; i<k; i++ ) if( X[i] < X[i-1] ) printf("NOT SORTED\n");
   return k;
}


LONG polnsparseeval64s(LONG *C, LONG *X, int t, LONG *A, int n, LONG p );
int vecmember64s( LONG x, LONG *S, int ns ) {
   // return true iff x==S[i] for 0<=i<ns
   int i;
   for( i=0; i<ns; i++ ) if( x==S[i] ) return 1;
   return 0;
}


int parevalakf0g0( int n,
                        LONG beta, int dAy, LONG *BETA,
                        LONG *A, LONG *X, int na, LONG *Ak, LONG *Xk, int *nAkn,
                        LONG *f, LONG *u, int nf, LONG *f0, int dfx, int dfy, int *dfn,
                        LONG *g, LONG *v, int ng, LONG *g0, int dgx, int dgy, int *dgn,
                        LONG p ) {
      evalpowers64s( beta, dAy, BETA, p ); // BETA[i] = beta^i mod p
      nAkn[n] = poltrivar3eval64s( A, X, na, BETA, Ak, Xk, p );        // (Ak,Xk) = (A,X)(y=beta)
      dfn[n]  = polbivareval64s(  f, u, nf, BETA, f0, dfx, dfy, p );  // f0(x) = (f,u)(y=beta)
      dgn[n]  = polbivareval64s(  g, v, ng, BETA, g0, dgx, dgy, p );  // g0(x) = (g,v)(y=beta)
      return 1;
}


int HenselLift2(
     LONG *A, LONG *X, int na, LONG *f, LONG *u, int nf, LONG *g, LONG *v, int ng, LONG alpha, LONG p, // inputs
     int NumCores,
     int **DegFi, LONG ***Fxyi, int **DegGi, LONG ***Gxyi, // temporary space
     int * DegFz, int *nF, LONG **F, LONG **Y, int * DegGz, int *nG, LONG **G, LONG **Z ) // outputs

// a(x,y,z) stored as (A,X) with na terms with monomials in X stored with [x>z>y] to facilitate evaluation at y
// f(x,y) stored in (f,u) with nf terms
// g(x,y) stored in (g,v) with ng terms
// a(x,y,alpha) = f(x,y) g(x,y) mod p
// (cF, xF, dFz) and (cG, xG, dGz) store the result
// Let F = sum( sig_i (z-alpha)^i, i=0..dF ) where sig_0 = f.
// Let G = sum( tau_i (z-alpha)^i, i=0..dG ) where tau_0 = g.
{
   LONG *Ak, *Xk, *f0, *g0;
   int *D;
   LONG *SPACE, *M, **U, *B, **BB; 
   LONG beta, *BETA, **BETAn; // space for evaluation at y=beta_i
   int i, j, k, N, df, dg, dAx, dAy, dAz, dfx, dfy, dgx, dgy, dFy, dGy, dFz, dGz, nAk, dx;
   int DD[3], *DegF, *DegG, failures, maxfails;
   LONG **Fxy, **Gxy, Mbeta;
   LONG **Fxzvals, **Gxzvals, *Omega, *OmegaInv; // for images of x in HenselLift1
   LONG **Akn, **Xkn, **f0n, **g0n, **SPACEn, **Fxzvalsn, **Gxzvalsn;
   int  n, NumImages, *nAkn, *dfn, *dgn;
   int ** Dn; LONG *** Un;
   recint P;
   double Hens1T, ST, NewtT, ConvT, DistT, Eval2T;

   if( NumCores<1 ) NumCores = 1; // fix this here rather than generating an error

   P = recip1(p);

   //polnsparsevecdeg64s( X, na, 3, DD ); dAx = DD[2]; dAy = DD[1]; dAz = DD[0];         
   // testing (d1,dj,d2) ordering to save sorting
   polnsparsevecdeg64s( X, na, 3, DD ); dAx = DD[2]; dAz = DD[1]; dAy = DD[0];         
   polnsparsevecdeg64s( u, nf, 2, DD ); dfx = DD[1]; dfy = DD[0];
   polnsparsevecdeg64s( v, ng, 2, DD ); dgx = DD[1]; dgy = DD[0]; 

   //printf("Hensel2:  na=%d  dAx=%d dAy=%d dAz=%d   dfy=%d dgy=%d  NumCores=%d\n",na,dAx,dAy,dAz,dfy,dgy,NumCores);

   //if( 1 ) { printf("  A := "); polnsparseprint64s(A, X, na, 3); printf(";\n"); }
   //if( 1 ) { printf("  f := "); polnsparseprint64s(f, u, nf, 2); printf(";\n"); }
   //if( 1 ) { printf("  g := "); polnsparseprint64s(g, v, ng, 2); printf(";\n"); }

   Akn = array2(NumCores,na); Ak = Akn[0]; // Ak = array(na);      
   Xkn = array2(NumCores,na); Xk = Xkn[0]; // Xk = array(na);      
   f0n = array2(NumCores,dfx+1); f0 = f0n[0]; // f0 = array(dfx+1);   
   g0n = array2(NumCores,dgx+1); g0 = g0n[0]; // g0 = array(dgx+1);   
   nAkn = arrayints(NumCores);
   dfn = arrayints(NumCores);
   dgn = arrayints(NumCores);

   n = 8*(dAx+1) + (dAz+1);
   // SPACE = array( n ); // space for polgcdex inside HenselLift1
   SPACEn = array2( NumCores, n ); SPACE = SPACEn[0];

   M = array(dAy+1); 
   BETAn = array2(NumCores,dAy+1); BETA = BETAn[0]; // BETA = array(dAy+1);

   //D = arrayints(dAy+1);
   //U = arrayofptrs(dAy+1);
   BB = array2(dAz+1,dAx+1); // B = array(dAx+1);

   Dn = array2ints(dAz+1,dAy+1);
   Un = (LONG ***) array2(dAz+1,dAy+1);

   N = fftgetn(dAx);
   if( (p-1)%N == 0 ) {
       LONG w, winv;
       Omega = array(N);        OmegaInv = array(N);
       w = getomega64s(p,N);    winv = modinv64s(w,p);
if( DEBUG ) printf("Using FFT for evaluations: N=%d\n",N);
       MakeW64(N,w,Omega,p);    MakeW64(N,winv,OmegaInv,p);
       // Fxzvals = array2(dAz+1,N); // space for evaluations of coefficients of x
       // Gxzvals = array2(dAz+1,N); // reuse these for each call to HenselLift1
       // Fxzvalsn = array2( NumCores*(dAz+1), N ); Fxzvals = Fxzvalsn;
       // Gxzvalsn = array2( NumCores*(dAz+1), N ); Gxzvals = Gxzvalsn;
       Fxzvalsn = array2( NumCores*N, dAz+1 ); Fxzvals = Fxzvalsn; /*TRANS*/
       Gxzvalsn = array2( NumCores*N, dAz+1 ); Gxzvals = Gxzvalsn; /*TRANS*/
   }
   else {
       Omega = NULL;
       OmegaInv = NULL;
       N = 0;
if( DEBUG ) printf("No FFT for evaluations\n");
       //Fxzvals = array2(dAz+1,dAx+1); // space for evaluations of coefficients of x
       //Gxzvals = array2(dAz+1,dAx+1); // reuse these for each call to HenselLift1
       Fxzvalsn = array2( NumCores*(dAz+1), dAx+1 ); Fxzvals = Fxzvalsn; /*TRANS*/
       Gxzvalsn = array2( NumCores*(dAz+1), dAx+1 ); Gxzvals = Gxzvalsn; /*TRANS*/
   }


   dFy = 0; dGy = 0;
   dFz = 0; dGz = 0;
   failures = 0;
   maxfails = 3;

   DiffT = 0.0;
   InterpT = 0.0;
   DioT = 0.0;
   NewtT = 0.0;
   ConvT = 0.0;
   DistT = 0.0;
   EvalT = 0.0;
   TaymulT = 0.0;
   Hens1T = 0.0;
   BevalT = 0.0;
   Eval2T = 0.0;

   i = 0;
   NumImages = 0;

   while( dFy + dGy < dAy && failures < maxfails ) {

      if( NumImages==0 ) {

          NumImages = min32s( NumCores, (dAy-dFy-dGy+1)/2 ); // must not go past dAy
          //printf(" Hensel2:  doing next %d images\n", NumImages );

          ST = realtime();
          for( n=0; n<NumImages; n++ ) { int dummy;
              beta = rand64s(p);
              while( beta==0 || vecmember64s(beta,M,i+n) ) beta = rand64s(p);
              M[i+n]=beta;
              Ak = Akn[n];
              Xk = Xkn[n];
              f0 = f0n[n];
              g0 = g0n[n];
              BETA = BETAn[n];
              // Evaluate A,f,g at y=beta
              // evalpowers64s( beta, dAy, BETA, p ); // BETA[i] = beta^i mod p
              // nAkn[n] = poltrivar3eval64s( A, X, na, BETA, Ak, Xk, p );        // (Ak,Xk) = (A,X)(y=beta)
              // dfn[n]  = polbivareval64s(  f, u, nf, BETA, f0, dfx, dfy, p );  // f0(x) = (f,u)(y=beta)
              // dgn[n]  = polbivareval64s(  g, v, ng, BETA, g0, dgx, dgy, p );  // g0(x) = (g,v)(y=beta)
              dummy = parevalakf0g0( n,
                        beta, dAy, BETA,
                        A, X, na, Ak, Xk, nAkn,
                        f, u, nf, f0, dfx, dfy, dfn,
                        g, v, ng, g0, dgx, dgy, dgn,
                        p );
          }
          ;
          Eval2T += realtime()-ST;


//for( n=0; n<NumImages; n++ ) {
//printf("beta = %lld\n",M[i+n]);
//printf("Ak[%d] = ",i+n); vecprint64s(Akn[n],nAkn[n]); printf(";\n");
//printf("Xk[%d] = ",i+n); vecprint64s(Xkn[n],nAkn[n]); printf(";\n");
//printf("f0[%d] = ",i+n); vecprint64s(f0n[n],dfn[0]+1); printf(";\n");
//printf("g0[%d] = ",i+n); vecprint64s(g0n[n],dgn[0]+1); printf(";\n");
//}

          ST = realtime();
          for( n=0; n<NumImages; n++ ) { int okay;
              Ak = Akn[n]; Xk = Xkn[n]; nAk = nAkn[n]; 
              f0 = f0n[n]; df = dfn[n];
              g0 = g0n[n]; dg = dgn[n];
              SPACE = SPACEn[n];
              DegF = DegFi[i+n]; Fxy = Fxyi[i+n];
              DegG = DegGi[i+n]; Gxy = Gxyi[i+n];
              // Fxzvals = Fxzvalsn + n*(dAz+1);
              // Gxzvals = Gxzvalsn + n*(dAz+1);
              Fxzvals = Fxzvalsn + n*N; /*TRANS*/
              Gxzvals = Gxzvalsn + n*N; /*TRANS*/
              okay =  HenselLift1( Ak, Xk, nAk, f0, df, g0, dg, alpha, p,  // inputs a(x1,xj), f0(x1), g0(x1)
                                         SPACE, Fxy, DegF, Gxy, DegG,            // space + outputs
                                         Fxzvals, Gxzvals, Omega, OmegaInv, N     // FFT space 
                                 );
          }
          ;
          Hens1T += realtime()-ST;
      }
      else {

          DegF = DegFi[i]; Fxy = Fxyi[i]; dFz = DegF[0];
          DegG = DegGi[i]; Gxy = Gxyi[i]; dGz = DegG[0];

          if( i==0 ) { i++; NumImages--; continue; }

          //printf("Hensel2: i=%d beta=%lld\n",i,M[i]);

          // Start Newton update (include new image)
          ST = realtime();
          Mbeta = 1; beta = M[i]; // M(y) = product( y-betaj, j=0..i-1 )
          for( j=0; j<i; j++ ) Mbeta = mulrec64(Mbeta,sub64s(beta,M[j],p),P);
          Mbeta = modinv64s(Mbeta,p); // Mbeta = M(beta)^(-1)

          for( j=0; j<=dFz; j++ ) {
             D = Dn[j];
             U = Un[j];
             B = BB[j];
             for( k=0; k<i; k++ ) { U[k] = Fxyi[k][j]; D[k] = DegFi[k][j+1]; }
//dx = cilkpolnewton2eval64s( i-1, D, U, M, beta, B, Fxy[j], Mbeta, DegF[j+1], p );
if(i>9)
DegF[j+1] = cilkpolnewton2eval64s( i-1, D, U, M, beta, B, Fxy[j], Mbeta, DegF[j+1], p );
else{ 
             dx = polnewton2eval64s( i-1, D, U, M, beta, B, p ); //printf(" B[%d,%d] := ",i,j); polprint64s(B,dx); printf(";\n");
             dx = polsub64s( Fxy[j], B, Fxy[j], DegF[j+1], dx, p );
             polscamul64s( Mbeta, Fxy[j], dx, p );
             DegF[j+1] = dx;
             //if( dx>=0 ) dFy = i; // printf(" V[%d,%d] := ",i,j); polprint64s(Fxy[j],dx); printf(";\n");
}
          }
if( i>10 )   ;

          for( j=0; j<=dFz; j++ ) { dx = DegF[j+1];
             if( dx>=0 ) dFy = i; // printf(" V[%d,%d] := ",i,j); polprint64s(Fxy[j],dx); printf(";\n");
          }

          for( j=0; j<=dGz; j++ ) {
             D = Dn[j];
             U = Un[j];
             B = BB[j];
             for( k=0; k<i; k++ ) { U[k] = Gxyi[k][j]; D[k] = DegGi[k][j+1]; }
//dx = cilkpolnewton2eval64s( i-1, D, U, M, beta, B, Gxy[j], Mbeta, DegG[j+1], p );
if(i>10)
DegG[j+1] = cilkpolnewton2eval64s( i-1, D, U, M, beta, B, Gxy[j], Mbeta, DegG[j+1], p );
else {
             dx = polnewton2eval64s( i-1, D, U, M, beta, B, p ); //printf(" B := "); polprint64s(B,dx); printf(";\n");
             dx = polsub64s( Gxy[j], B, Gxy[j], DegG[j+1], dx, p );
             polscamul64s( Mbeta, Gxy[j], dx, p );
             DegG[j+1] = dx;
}
             //if( dx>=0 ) dGy = i; // printf(" V[%d,%d] := ",i,j); polprint64s(Gxy[j],dx); printf(";\n");
          }
if( i>10)           ;
          for( j=0; j<=dGz; j++ ) { dx = DegG[j+1];
             if( dx>=0 ) dGy = i; // printf(" V[%d,%d] := ",i,j); polprint64s(Fxy[j],dx); printf(";\n");
          }

          NewtT += realtime()-ST;
          NumImages --; i++;
      }
 
/*
      beta = rand64s(p); while( beta==0 || vecmember64s(beta,M,i) ) beta = rand64s(p);
      M[i]=beta;

      DegF = DegFi[i]; Fxy = Fxyi[i];
      DegG = DegGi[i]; Gxy = Gxyi[i];


          printf("Hensel2: i=%d beta=%lld\n",i,beta);

      // Evaluate A,f,g at y=beta
      evalpowers64s( beta, dAy, BETA, p ); // BETA[i] = beta^i mod p
      nAk = poltrivareval64s( A, X, na, BETA, Ak, Xk, p );        // (Ak,Xk) = (A,X)(y=beta)
      df  = polbivareval64s(  f, u, nf, BETA, f0, dfx, dfy, p );  // f0(x) = (f,u)(y=beta)
      dg  = polbivareval64s(  g, v, ng, BETA, g0, dgx, dgy, p );  // g0(x) = (g,v)(y=beta)

          //printf("Ak := "); polnsparseprint64s(Ak, Xk, nAk, 2); printf(";\n");
          //printf("f0 := "); polprint64s(f0,df); printf(";\n");
          //printf("g0 := "); polprint64s(g0,dg); printf(";\n");

          ST = realtime();

      if( !HenselLift1( Ak, Xk, nAk, f0, df, g0, dg, alpha, p,  // inputs a(x1,xj), f0(x1), g0(x1)
                        SPACE, Fxy, DegF, Gxy, DegG,            // space + outputs
                        Fxzvals, Gxzvals, Omega, OmegaInv, N )  // FFT space 
                      ) { failures ++; continue; }

           Hens1T += realtime()-ST;

	//printf("Hensel2: i=%d, beta=%lld  DegF[0] = %d\n",i,beta,DegF[0]);
	//         printf("fk := "); pol2tayprint64s( DegF[0], DegF+1, Fxy, alpha ); printf(";\n");
	//         printf("gk := "); pol2tayprint64s( DegG[0], DegG+1, Fxy, alpha ); printf(";\n");
	//for( j=0; j<=DegFi[i][0]; j++ ) printf("DEG:  %d ",DegFi[i][j]); printf("\n");
	//for( j=0; j<=DegFi[i][0]; j++ )
	//   if( i==0 ) { printf("V[%d,%d] := ",i,j); polprint64s(Fxyi[i][j],DegFi[i][j+1]); printf(";\n"); }


      // Interpolate y in Fxy, Gxy
      dFz = DegF[0];
      dGz = DegG[0];

      if( i==0 ) { i++; continue; }

      // Start Newton update (include new image)
      ST = realtime();
      Mbeta = 1; // M(y) = product( y-betaj, j=0..i-1 )
      for( j=0; j<i; j++ ) Mbeta = mulrec64(Mbeta,sub64s(beta,M[j],p),P);
      Mbeta = modinv64s(Mbeta,p); // Mbeta = M(beta)^(-1)

      for( j=0; j<=dFz; j++ ) {
         for( k=0; k<i; k++ ) { U[k] = Fxyi[k][j]; D[k] = DegFi[k][j+1]; }
         dx = polnewton2eval64s( i-1, D, U, M, beta, B, p ); //printf(" B[%d,%d] := ",i,j); polprint64s(B,dx); printf(";\n");
         dx = polsub64s( Fxy[j], B, Fxy[j], DegF[j+1], dx, p );
         polscamul64s( Mbeta, Fxy[j], dx, p );
         DegF[j+1] = dx;
         if( dx>=0 ) dFy = i; // printf(" V[%d,%d] := ",i,j); polprint64s(Fxy[j],dx); printf(";\n");
      }
      for( j=0; j<=dGz; j++ ) {
         for( k=0; k<i; k++ ) { U[k] = Gxyi[k][j]; D[k] = DegGi[k][j+1]; }
         dx = polnewton2eval64s( i-1, D, U, M, beta, B, p ); //printf(" B := "); polprint64s(B,dx); printf(";\n");
         dx = polsub64s( Gxy[j], B, Gxy[j], DegG[j+1], dx, p );
         polscamul64s( Mbeta, Gxy[j], dx, p );
         DegG[j+1] = dx;
         if( dx>=0 ) dGy = i; // printf(" V[%d,%d] := ",i,j); polprint64s(Gxy[j],dx); printf(";\n");
      }
      NewtT += realtime()-ST;
      i++;
*/

   } // END WHILE

   if( failures==maxfails ) { printf("Hensel2: Hensel 1 failed %d times.\n",maxfails); return 0; }

   DegFz[0] = dFz;
   DegGz[0] = dGz;

   ST = realtime();

{  int *DD; int dummy; LONG **UU; int k;


   // convert coefficients in y from newton form to standard form
   for( j=0; j<=dFz; j++ ) {
         DD = Dn[j]; UU = Un[j];
         for( k=0; k<=dFy; k++ ) { UU[k] = Fxyi[k][j]; DD[k] = DegFi[k][j+1]; }
         if( dFy>19 ) { dummy = cilknewton2standard64s( dFy, DD, UU, M, p );
                        if( (j+1)%NumCores == 0 || j==dFz ) ; }
         else newton2standard64s( dFy, DD, UU, M, p );
   }
   //sync;

   for( j=0; j<=dFz; j++ ) {
         DD = Dn[j];
         for( k=0; k<=dFy; k++ ) { DegFi[k][j+1] = DD[k]; }
   }

   for( j=0; j<=dGz; j++ ) {
         DD = Dn[j]; UU = Un[j];
         for( k=0; k<=dGy; k++ ) { UU[k] = Gxyi[k][j]; DD[k] = DegGi[k][j+1]; }
         if( dGy>19 ) { dummy = cilknewton2standard64s( dGy, DD, UU, M, p );
                        if( (j+1)%NumCores == 0 || j==dGz ) ; }
         else newton2standard64s( dGy, DD, UU, M, p );
   }
   //sync;

   for( j=0; j<=dGz; j++ ) {
         DD = Dn[j];
         for( k=0; k<=dGy; k++ ) { DegGi[k][j+1] = DD[k]; }
   }


}
   ConvT = realtime()-ST;

   //printf("Newton interpolation of x2 done.\n");

   // We have F = sum( sum( f_ij(x) y^j, j=0..dFy) (z-alpha)^i, i=0..dFz )
   // and     G = sum( sum( g_ij(x) y^j, j=0..dGy) (z-alpha)^i, i=0..dGz )

   ST = realtime();
   for( j=0; j<=dFz; j++ ) {
      //printf("dFz=%d: j=%d\n",dFz,j);
      for( k=0; k<=dFy; k++ ) { U[k] = Fxyi[k][j]; D[k] = DegFi[k][j+1]; }
      nF[j] = rec2bivar64s( dFy, D, U, F[j], Y[j] );
      //if( j>0 ) printf("\n +"); else printf("FF := ");
      //printf("("); polnsparseprint64s(F[j],Y[j],nF[j],2); printf(")*(z-%lld)^%d",alpha,j);
   }
      //printf(";\n");

   for( j=0; j<=dGz; j++ ) {
      for( k=0; k<=dGy; k++ ) { U[k] = Gxyi[k][j]; D[k] = DegGi[k][j+1]; }
      nG[j] = rec2bivar64s( dGy, D, U, G[j], Z[j] );
      //if( j>0 ) printf("\n +"); else printf("GG := ");
      //printf("("); polnsparseprint64s(G[j],Z[j],nG[j],2); printf(")*(z-%lld)^%d",alpha,j); 
   }
      //printf(";\n");
   DistT = realtime()-ST;

//printf("Hensel2 done:  dFy=%d  dGy=%d  dAy=%d\n",dFy,dGy,dAy);
//printf("Hens1time=%10.4f  DiffT=%10.4f  BevalT=%10.4f  Diotime=%10.4f\n",Hens1T,DiffT,BevalT,DioT);
//printf("    EvalT=%10.4f  TaymulT=%10.4f  InterpT=%10.4f  \n",EvalT,TaymulT,InterpT);
//printf("Hens2time:  Hens1T=%10.4f  Eval2T=%9.4f  NewtT=%9.4f  ConvT=%9.4f  DistT=%9.4f\n",Hens1T,Eval2T,NewtT,ConvT,DistT);


   // Now we have F = sum( F_i(y,x) (z-alpha)^i, i=0..dFz )
   // and we have G = sum( G_i(y,x) (z-alpha)^i, i=0..dGz )
   // Note monomials are in lex with y>x not x>y because they are sorted in this order

if( 0 ) { int zero;

   printf("p := %lld;\n",p);

   printf("A := ");
   polnsparseprint64s(A, X, na, 3);
   printf(";\n");

   printf("F := ");
   for( j=0; j<=dFz; j++ ) {
      zero = 1;
      for( k=0; k<=dFy; k++ ) { 
          U[k] = Fxyi[k][j];
          D[k] = DegFi[k][j+1];
          if( D[k]>=0 ) zero = 0;
      }
      if( zero ) { if( j==0 ) printf("0"); else printf("\n+0*z^%d",j); continue; }
      if( j>0 ) printf("\n+("); else printf("(");
      for( k=0; k<=dFy; k++ )
          if( D[k]>=0 ) { 
              if(k==0) printf("("); else printf("+(");
              polprint64s(U[k],D[k]);
              printf(")*y^%d",k); }
          else if(k==0) printf("0");
      //printf(")*(z-%lld)^%d",alpha,j);
      printf(")*z^%d",j);
   }
   printf(";\n");

   printf("G := ");
   for( j=0; j<=dGz; j++ ) {
      zero = 1;
      for( k=0; k<=dGy; k++ ) {
          U[k] = Gxyi[k][j];
          D[k] = DegGi[k][j+1]; 
          if( D[k]>=0 ) zero = 0;
      }
      if( zero ) { if( j==0 ) printf("0"); else printf("\n+0*z^%d",j); continue; }
      if( j>0 ) printf("\n+("); else printf("(");
      for( k=0; k<=dGy; k++ )
          if( D[k]>=0 ) { 
              if(k==0) printf("("); else printf("+(");
              polprint64s(U[k],D[k]);
              printf(")*y^%d",k); }
          else if(k==0) printf("0");
      //printf(")*(z-%lld)^%d",alpha,j);
      printf(")*z^%d",j);
   }
   printf(";\n");

}

   freemem( Fxzvalsn[0] ); freemem( Fxzvalsn ); //freemem( Fxzvals[0] ); freemem( Fxzvals );
   freemem( Gxzvalsn[0] ); freemem( Gxzvalsn ); //freemem( Gxzvals[0] ); freemem( Gxzvals );
   freemem( SPACEn[0] ); freemem( SPACEn ); // freemem( SPACE );
   freemem( M );
   freemem(BETAn[0] ); freemem( BETAn ); // freemem( BETA );
   freemem(Dn[0]); freemem(Dn);  // freemem( D );
   freemem(Un[0]); freemem(Un);  // freemem( U );
   freemem( BB[0] );  freemem( BB );  // freemem( B );
   freemem( Akn[0] ); freemem( Akn ); // freemem( Ak );
   freemem( Xkn[0] ); freemem( Xkn ); // freemem( Xk );
   freemem( f0n[0] ); freemem( f0n ); // freemem( f0 );
   freemem( g0n[0] ); freemem( g0n ); // freemem( g0 );
   freemem( nAkn );
   freemem( dfn );
   freemem( dgn );
   if( Omega!=NULL ) freemem(Omega);
   if( OmegaInv!=NULL ) freemem(OmegaInv);

   return 1;

}

void monexpand1(LONG *X, int na, int n) 
{
   LONG x, y, mask;
   int oldbits,newbits,i,j;
   oldbits=63/n;
   mask = (1LL<<oldbits)-1;
   newbits=63/(n+1);
   for( j=0; j<na; j++ ) {
      x = X[j];
      y = 0;
      for( i=0; i<n; i++ ) {
          y = ((x & mask) << (i*newbits)) | y;
          x = x >> oldbits;
      }
      X[j] = y << newbits;
   }
}


int polnsparseadd64s( LONG *A, LONG *X, int na, LONG *B, LONG *Y, int nb, LONG *C, LONG *Z, LONG p )
{  // (C,Z) = (A,X) + (B,Y) using a merge X,Y,Z are monomials, A,B,C are coefficients mod p
   int i,j,k;
   LONG c;
   for( i=0,j=0,k=0; i<na && j<nb; ) {
      if( X[i]<Y[j] ) { C[k] = A[i]; Z[k] = X[i]; i++; k++; }
      else if( Y[j]<X[i] ) { C[k] = B[j]; Z[k] = Y[j]; j++; k++; }
      else { c = add64s(A[i],B[j],p); if( c!=0 ) { C[k] = c; Z[k] = X[i]; k++; } i++; j++; }
   }
   // remaining terms
   while( i<na ) { C[k] = A[i]; Z[k] = X[i]; i++; k++; }
   while( j<nb ) { C[k] = B[j]; Z[k] = Y[j]; j++; k++; }
   return k;
}


int hornerstep( LONG *A, LONG *X, int na, LONG *B, LONG *Y, LONG alpha, LONG p ) {
   // compute b = (z-alpha) a
   // the multiplication z a is done by adding 1 to the monomials in X
   int i,j,k; LONG c; recint P;
   P = recip1(p);
   for( i=0,j=0,k=0; i<na && j<na; ) {
       if( X[i]+1 < X[j] ) { B[k] = A[i]; Y[k] = X[i]+1; i++; k++; }
       else if( X[j] < X[i]+1 ) { B[k] = neg64s( mulrec64(alpha,A[j],P), p ); Y[k] = X[j]; j++; k++; }
       else { c = sub64s( A[i], mulrec64(alpha,A[j],P), p ); 
              if( c!=0 ) { B[k] = c; Y[k] = X[j]; k++; }
              i++; j++; 
            }
   }
   while( i<na ) { B[k] = A[i]; Y[k] = X[i]+1; i++; k++; }
   while( j<na ) { B[k] = neg64s(mulrec64(alpha,A[j],P),p); Y[k] = X[j]; j++; k++; }
   return k;
}


int HornerExpand( int d, LONG alpha, int n, int *N, LONG **F, LONG **Y, LONG *A, LONG *X, LONG *T1, LONG *T2, LONG p ) 
{  // A = expand( F0(x1,...,xn) + (z-alpha)( F1(x1,...,xn) + (z-alpha)( F2(x1,...,xn) + (z-alpha)( F3(x1,...,xn) ))) ) 
   // The result in (A,X) needs to be as large as (d+1)F
   // Because the monomials in Y[i] are a subset of those in Y[i-1], (A,X) has at most at (d+1) N[0] terms.
   // Similarly (T1,T2) has at most (d+1) N[0] terms.
   int i,na,nt;
   na = N[d];
   polcopy64s( F[d], na, A );
   monexpand1( Y[d], na, n );  // compress x1,x2,...,xn to make space for z as xn+1
   polcopy64s( Y[d], na, X );
   for( i=d-1; i>=0; i-- ) {
       nt = hornerstep( A, X, na, T1, T2, alpha, p ); // T = (z-alpha) A
       monexpand1( Y[i], N[i], n ); // make space for xn+1
       na = polnsparseadd64s( F[i], Y[i], N[i], T1, T2, nt, A, X, p ); // A = F[i] + T
       //printf("   TT := "); polnsparseprint64s( A, X, na, n+1 ); printf(";\n");
   }
   return na;
}

int Expand( int d, int n, int *N, LONG **F, LONG **Y, LONG *A, LONG *X, LONG p ) 
{  // A = expand( F0(x1,...,xn) + F1(x1,...,xn) z + F2(x1,...,xn) z^2 + F3(x1,...,xn) z^d )
   int i,j,na;
   na = 0;
   //printf("Expand: d=%d n=%d\n",d,n);
   for( i=0; i<=d; i++ ) {
       monexpand1( Y[i], N[i], n );
       for( j=0; j<N[i]; j++ ) {
           A[na+j] = F[i][j];
           X[na+j] = Y[i][j] + i; // * z^i
       }
       na += N[i];
    }
    //printf(" A := "); polnsparseprint64s( A, X, na, n+1 ); printf(";\n");
    return na;
}


LONG polbinsearch64s( LONG *A, LONG *X, int na, LONG mon ) 
{  // search for mon in X and return the corresponding coefficient in A
   int mid,top,bot;
   bot = 0;
   top = na-1;
   while( top >= bot ) {
       mid = (bot+top)/2;
       if( mon < X[mid] ) top = mid-1;
       else if (mon > X[mid]) bot = mid+1;
       else return A[mid];
   }
   return 0;
}
