// Copyright, Michael Monagan, 2020
// This file was created by making a copy of mtshl5.cilk stripping out the
// Cilk directives spawn, sync, cilk using  s/spawn//  s/sync//  s/cilk / /


#include <stdio.h>

#define DEBUG 0

#define LONG long long int

#define add64s add64seval
#define sub64s sub64seval
#define neg64s neg64seval
#define mul64s mul64seval
#define max32s max32seval
#define min32s min32seval

LONG add64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
LONG sub64s(LONG a, LONG b, LONG p) { LONG t; t = a-b; t += (t>>63) & p; return t; }
LONG neg64s(LONG a, LONG p) { LONG t; t = -a; t += (t>>63) & p; return t; }
int max32s(int a, int b) { if( a<b ) return b; else return a; }
int min32s(int a, int b) { if( a<b ) return a; else return b; }
LONG mul64s(LONG a, LONG b, LONG p) {
        LONG q, r;
        __asm__ __volatile__(           \
        "       mulq    %%rdx           \n\t" \
        "       divq    %4              \n\t" \
        : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
        return r;
}


LONG modinv64s( LONG c, LONG p );
LONG powmod64s( LONG a, LONG n, LONG p );

#define UINT64 unsigned long long
typedef struct {
	UINT64 s;	/* shift */
	UINT64 v;	/* reciprocal of d */
	UINT64 d0;	/* divisor shifted up */
	UINT64 d1;
} recint;

recint recip1(UINT64  p);
UINT64 mulrec64(UINT64  a, UINT64  b, recint  v);

void sort2vec64s(LONG *D, LONG *C, int n );
void vecprint64s( LONG *A, int n );
void vecprint32s( int *A, int n );

double realtime();

void evalbeta( LONG *A, int d, LONG beta, recint P ) {
//  A[i] = beta^i mod p
    int i;
    if( d<0 ) return;
    A[0] = 1;
    if( d==0 ) return; 
    A[1] = beta;
    for( i=2; i<=d; i++ ) A[i] = mulrec64(beta,A[i-1],P);
    return;
}


int polneval1(LONG *A, LONG *X, int t, LONG *ALPHA, LONG *B, LONG *Y, int n, LONG p ) {
// Evaluate f(x1,x2,...,xn1,xj=alpha) 
    int numbits,numbitsB,i,j,k;
    LONG mask,x,y,d,c;
    recint P;
    P = recip1(p);
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    numbitsB = 63/(n-1);
    for( i=0; i<t; i++ ) {
        x = X[i]; // old monomial
        d = x & mask;
        B[i] = mulrec64(A[i],ALPHA[d],P);
        y = 0;
        x = x >> numbits;
        for( j=0; j<n-1; j++ ) {
            d = x & mask;
            y = y | (d << j*numbitsB );
            x = x >> numbits;
        }
        Y[i] = y; // new monomial
    }
    for( k=0,i=0; i<t; i=j ) { // add up coefficients of equal monomials
        c = B[i];
        for( j=i+1; j<t && Y[i]==Y[j]; j++ ) c = add64s(c,B[j],p);
        if( c!=0 ) { B[k] = c; Y[k] = Y[i]; k++; }
    }
    return k;
}


void polnsparseprint64s( LONG *C, LONG *X, int t, int n );


void evalhens( LONG *A, LONG *X, int t, int n ) {
    // X is updated:  (d1 d2 d3 d4 d5 ... dn) ==> (d1 d2 dn d3 d4 d5 ...) 
    int i,numbits;
    LONG maskbot, maskmid, x, y, xn, xmid, x1x2;
    numbits = 63/n;
    maskbot = (((LONG) 1) << numbits) - 1 ;
    maskmid = (((LONG) 1) << (n-3)*numbits) - 1;
    for( i=0; i<t; i++ ) {
       x = X[i];
       xn = (x & maskbot) << (n-3)*numbits;
       x1x2 = x >> (n-2)*numbits;
       x1x2 = x1x2 << (n-2)*numbits;
       xmid = (x >> numbits) & maskmid;
       y = (x1x2 | xn | xmid);
       X[i] = y;
    }
    return;
}


void reorderxn( LONG *A, LONG *X, int t, int n ) {
    // X is updated:  (d1 d2 d3 ... dn) ==> (d1 dn d2 d3 ...)
    int i,numbits;
    LONG maskbot, maskmid, x, y, x1, xmid, xn;
    numbits = 63/n;
    maskbot = (((LONG) 1) << numbits) - 1 ;
    maskmid = (((LONG) 1) << (n-2)*numbits) - 1;
    for( i=0; i<t; i++ ) {
       x = X[i];
       xn = (x & maskbot) << (n-2)*numbits;
       x1 = (x >> (n-1)*numbits) << (n-1)*numbits;
       xmid = (x >> numbits) & maskmid;
       y = (x1 | xn | xmid);
       X[i] = y;
    }
    return;
}


double SORTTIME;

int vec2partition64s(LONG *D, LONG *C, int n );
int cilksort64s( LONG *A, LONG *X, int n ) {
    int i, j;
    if( n<200000 ) { sort2vec64s( X, A, n ); return 1; }
    i = vec2partition64s( X, A, n );
    j =  cilksort64s( A, X, i );
    j =  cilksort64s( A+i+1, X+i+1, n-i-1 );
    
    return j;
}


void evalinit( LONG *X, int na, int n, int xi, int xj, LONG **BETA, LONG *B, LONG *Y, LONG p )
{   
    // evaluate variables xi .. xj at BETA in (A,X) and return new polynomial in (B,Y)
    // 1 <= xi <= xj <= xn
    // there could be terms which evaluate to the same monomial note
    int i,j,m,numbitsA,numbitsB;
    LONG maskA,maskB,x,y,c,d;
    recint P;
    P = recip1(p);

    xi--; xj--;  // 0 <= xi <= xj < xn
    numbitsA = 63/n;
    maskA = (((LONG) 1) << numbitsA) - 1 ;
    m = n-(xj-xi)-1; 
    numbitsB = 63/m;
    maskB = (((LONG) 1) << numbitsB) - 1 ;

    for( i=0; i<na; i++ ) {
        x = X[i];
        c = 1;
        y = 0;
        for( j=n-1; j>=0; j-- ) {
            d = x & maskA;
            if( j < xi || j > xj ) y = (y << numbitsB) | d; else c = mulrec64(c,BETA[j-xi][d],P);
            x = x >> numbitsA;
        }
        B[i] = c;
        // reverse the bit fields in y (they are backwards)
        x = 0;
        for( j=0; j<m; j++ ) { d = y & maskB; x = (x << numbitsB) | d; y = y >> numbitsB; }
        Y[i] = x;
    }
    return;
}


int cilkevalinit( LONG *X, int na, int n, int xi, int xj, LONG **BETA, LONG *B, LONG *Y, LONG p )
{ evalinit( X, na, n, xi, xj, BETA, B, Y, p ); return 1; }


int cilkinitpar( int NumCores, LONG *X, int na, int n, int xi, int xj, LONG **BETA, LONG *B, LONG *Y, LONG p )
{    int i, m, j;
     if( NumCores<2 || na<1000000 ) { evalinit( X, na, n, xi, xj, BETA, B, Y, p ); return 1; }
     m = na/NumCores;
     for( i=0; i<NumCores-1; i++ ) j =  cilkevalinit( X+i*m, m, n, xi, xj, BETA, B+i*m, Y+i*m, p );
     evalinit( X+i*m, na-i*m, n, xi, xj, BETA, B+i*m, Y+i*m, p );
     
     return 1;
}


void  VECMULINP64s( LONG *A, LONG *G, int na, LONG p );


int evalnext( LONG *A, LONG *X, LONG *G, int na, LONG *B, LONG *Y, LONG p )
{   int i,j,k,n;
    LONG c,t,x;
    recint P;
    P = recip1(p);
    for( i=0,k=0; i<na; i=j ) {
        x = X[i];
        for( n=1,j=i+1; j<na && X[j]==x; j++,n++ );
        c = 0;
        for( j=i; n>0; j++,n-- ) {
            t = A[j];
            t = mulrec64(t,G[j],P); c = add64s(t,c,p);
            A[j] = t;
        }
        if( c!=0 ) { B[k] = c; Y[k] = x; k++; }
    }
    return k;
}


int cilkevalnext( LONG *A, LONG *X, LONG *G, int na, LONG *B, LONG *Y, LONG p )
{ return evalnext( A, X, G, na, B, Y, p ); }


// I1[i] must be at least size (dax+1)*(day+1)*(daz+1)
int cilkevalpar( LONG *A, LONG *X, LONG *G, int na, LONG **I1, LONG **Y1, int NumCores, LONG p ) 
{    int i, j, m, t, nB[64];
     LONG c, *B, *Y, *C, *Z;
     if( NumCores<2 || na<1000000 ) return evalnext( A, X, G, na, I1[0], Y1[0], p );
     m = na/NumCores;
     for( i=0; i<NumCores-1; i++ ) 
        nB[i] =  cilkevalnext( A+i*m, X+i*m, G+i*m, m, I1[i], Y1[i], p ); 
     nB[i] = evalnext( A+i*m, X+i*m, G+i*m, na-i*m, I1[i], Y1[i], p ); 
     

     printf("\ncilkevalpar: #cores=%d  m=%d  nB =",NumCores,m);
     for( i=0; i<NumCores; i++ ) printf("  %d",nB[i]); printf("\n");
     B = I1[0]; Y = Y1[0]; t = nB[0];
     for( i=1; i<NumCores; i++ ) {
         C = I1[i]; Z = Y1[i]; j = 0;
         if( Y[t-1] == Z[0] ) { c = add64s(B[t-1],C[0],p); if( c!=0 ) { B[t-1] = c; t++; } j = 1; }
         for( ; j<nB[i]; j++ ) { B[t] = C[j]; Y[t] = Z[j]; t++; }
     }
     return t;
}


LONG packlong( int n, int m ) { return( (((LONG) n) << 32) + m ); }

int unpacklong( LONG n, int i ) {
   LONG mask; 
   mask = ((1LL) << 32) - 1;
   if( i==1 ) return (n >> 32); 
   if( i==2 ) return (n & mask);
   printf("ERROR in unpacklong\n");
   return 0;
}


LONG evalnext2( LONG *A, LONG *X, LONG *G, int na, LONG *B, LONG *C, LONG *Y, LONG *Z, LONG p, int MAX );
LONG cilkevalnext2( LONG *A, LONG *X, LONG *G, int na, LONG *B, LONG *C, LONG *Y, LONG *Z, LONG p, int MAX )
{ return evalnext2( A, X, G, na, B, C, Y, Z, p, MAX ); }


// the arrays I1[i], I2[i], Y1[i], Y2[i] must be at least size (dax+1)*(day+1)*(daz+1)
LONG cilkevalpar2( LONG *A, LONG *X, LONG *G, int na, LONG **I1, LONG **I2, LONG **Y1, LONG **Y2, int NumCores, LONG p, int MAX ) 
{    int i, j, m, s, t, n1[64], n2[64];
     LONG *B, *Y, *C, *Z, nB[64];

     if( NumCores<2 || na<100000 ) return evalnext2( A, X, G, na, I1[0], I2[0], Y1[0], Y2[0], p, MAX );

     m = na/NumCores;
     for( i=0; i<NumCores-1; i++ ) nB[i] =  cilkevalnext2( A+i*m, X+i*m, G+i*m, m, I1[i], I2[i], Y1[i], Y2[i], p, MAX ); 
     nB[i] = evalnext2( A+i*m, X+i*m, G+i*m, na-i*m, I1[i], I2[i], Y1[i], Y2[i], p, MAX ); 
     

     for( i=0; i<NumCores; i++ ) { n1[i] = unpacklong(nB[i],1); n2[i] = unpacklong(nB[i],2); }
     //printf("\ncilkevalpar2: #cores=%d  m=%d  nB =",NumCores,m);
     //for( i=0; i<NumCores; i++ ) printf("  (%d,%d)",n1[i],n2[i]); printf("\n");

     B = I1[0]; Y = Y1[0]; t = n1[0];
     for( i=1; i<NumCores; i++ ) {
         C = I1[i]; Z = Y1[i]; j = 0;
         if( Y[t-1] == Z[0] ) { B[t-1] = add64s(B[t-1],C[0],p); j = 1; }
         for( ; j<n1[i]; j++ ) { B[t] = C[j]; Y[t] = Z[j]; t++; }
     }
     s = t;
     B = I2[0]; Y = Y2[0]; t = n2[0];
     for( i=1; i<NumCores; i++ ) {
         C = I2[i]; Z = Y2[i]; j = 0;
         if( Y[t-1] == Z[0] ) { B[t-1] = add64s(B[t-1],C[0],p); j = 1; }
         for( ; j<n2[i]; j++ ) { B[t] = C[j]; Y[t] = Z[j]; t++; }
     }
     if( DEBUG) if( t>MAX ) printf(" t=%d is OUTOFRANGE of m=%d\n",t,m);
     return packlong(s,t); // hack to return two values
}


LONG * array( int n );
LONG ** arrayofptrs( int n );
int * arrayints( int n );
int ** arrayptrsints( int n );
LONG **array2( int n, int m );
int **array2ints( int n, int m );
void freemem(void *p);


void printsupp( LONG *X, int t, int n ) {
    int numbits,i,j,d1,d2,d; LONG mask,x;
    numbits = 63/n;
    mask = (((LONG) 1) << numbits) - 1 ;
    for( i=0; i<t; i++ ) {
        x = X[i];
        d1 = x >> ( (n-1)*numbits );
        d2 = ( x >> ( (n-2)*numbits ) ) & mask;
        printf(" x1^%d x2^%d : ", d1, d2 );
        for( j=0; j<n-2; j++ ) {
            d = x & mask;
            printf(" x%d^%d ",n-j,d );
            x = x >> numbits;
        }
        printf("\n");
    }
}


int monsortedcheck( LONG *X, int t ) {
    int i;
    for( i=0; i<t-1; i++ ) if( X[i]>X[i+1] ) return 0;
    return 1;
}
 

int suppsizes( LONG *X, int t, int n, int k, int *N ) {
    // monomials in X are in x1 ... xn.  compute support sizes in x1 ... xk
    int numbits,shift,i,j,s,m;
    if( DEBUG ) if( !monsortedcheck(X,t) ) printf("  suppsizes: X IS NOT SORTED\n");
    numbits = 63/n;
    shift = (n-k)*numbits;
    m = 0;
    for( i=0; i<t; i=j ) {
       for( s=1, j=i+1; j<t && (X[i]>>shift) == (X[j]>>shift); j++, s++ );
       N[m] = s; m++;
    }
    return m; // number of non-empty supports in x1,x2,...,xk
}


int distinctsupp( LONG *X, LONG *G, int t, int n, int k ) {
    int numbits,shift,i,j,h,s,distinct;
    numbits = 63/n;
    shift = (n-k)*numbits;
    distinct = 1;
//printf("distinctsupp:  n=%d  k=%d  shift=%d\n",n,k,shift);
//printf("    X: ["); for( i=0; i<t; i++ ) printf("%lld, ",X[i]>>shift); printf("]\n");
//printf("    G:"); vecprint64s(G,t);
    for( i=0; i<t && distinct; i=j ) {
       for( s=1, j=i+1; j<t && (X[i]>>shift) == (X[j]>>shift); j++, s++ );
       // check if G[i],G[i+1],...,G[i+s-1] are distinct
       for( s--; s>0 && distinct; s-- )
           for( h=0; h<s; h++ ) distinct = distinct && (G[i+h]!=G[i+s]);
    }
    return distinct;
}

 
void treemul( LONG *M, int t, LONG *L, LONG *T, LONG p );
void getlambda( LONG *M, int t, LONG *L, LONG p ) {
// compute (x-M[1])(x-M[2])(x-M[3])...(x-M[t]) vi repeated multiplication for t >= 0
    int i,j;
    LONG alpha;
    recint P;
    P = recip1(p);
    L[0] = 1;
    for( i=0; i<t; i++ ) {
        alpha = M[i];
        L[i+1] = 1;
        for( j=i; j>0; j-- ) L[j] = sub64s(L[j-1],mulrec64(alpha,L[j],P),p);
        L[0] = neg64s(mulrec64(alpha,L[0],P),p);
    }
    return;
}


void polnsparseprint64s( LONG *C, LONG *X, int t, int n );
void taylorprint64s( int dz, LONG alpha, LONG **A, LONG **X, int *N, int n )
{
    int i;
    for( i=0; i<=dz; i++ ) {
        if( i>0 ) printf("+");
        printf("(");
        polnsparseprint64s( A[i], X[i], N[i], n );
        printf(")");
        if( i!=0 ) printf("*(z-%lld)^%d",alpha,i);
    }
   return;
}


LONG rand64s(LONG p);
LONG polnsparseeval64s(LONG *C, LONG *X, int t, LONG *A, int n, LONG p );


int  henselcheck(LONG *A, LONG *X, int na, int n, 
                 int dFz, LONG **FC, LONG **FX, int *nF,
                 int dGz, LONG **GC, LONG **GX, int *nG,
                 LONG alpha, LONG p )
// Test if A(x1,...,xn) = sum( F[i] (z-alpha)^i, i=0..dFz ) sum( G[i] (z-alpha)^i, i=0..dGz ) at a random point beta1,...,betan mod p
{   
    int i; LONG Aval,Zval,Fval,Gval,beta[32],temp;
    if( n>32 ) { printf("Hensel Check: too many variables\n"); }
    for( i=0; i<n; i++ ) beta[i] = rand64s(p);
    Aval = polnsparseeval64s(A, X, na, beta, n, p );

if( n==3 ) // Sorry, in HenselLift2 for monomials [x1>x2>x3] I switched the output order to [x2>x1] to avoid an explicit sort
           // and in MTSHL I switched the order from [x1>x2>...>xj] to be [x1>xj>x2>...] to avoid another explicit sort
           // Also, polnsparseeval64s evaluates f(x1,x2,x3) as f(beta[2],beta[1],beta[0]) ..... another challenge!
           //{ temp = beta[1]; beta[1] = beta[2]; beta[2] = temp; } 
           // permute [beta[0],beta[1],beta[2]] to [beta[1],beta[2],beta[0]]
           { temp = beta[1]; beta[1] = beta[2]; beta[2] = beta[0]; beta[0] = temp; } 

    Zval = sub64s(beta[0],alpha,p);
    Fval = polnsparseeval64s(FC[dFz],FX[dFz],nF[dFz],beta+1,n-1,p);
    for( i=dFz-1; i>=0; i-- ) Fval = add64s( polnsparseeval64s(FC[i],FX[i],nF[i],beta+1,n-1,p), mul64s(Fval,Zval,p), p );
    Gval = polnsparseeval64s(GC[dGz],GX[dGz],nG[dGz],beta+1,n-1,p);
    for( i=dGz-1; i>=0; i-- ) Gval = add64s( polnsparseeval64s(GC[i],GX[i],nG[i],beta+1,n-1,p), mul64s(Gval,Zval,p), p );
    temp = sub64s(Aval,mul64s(Fval,Gval,p),p);
    if( temp!=0 ) {
        printf("In henselcheck n=%d  na=%d  dfz=%d  dgz=%d.\n",n,na,dFz,dGz);
        printf("   p := %lld;  alpha := %lld;\n",p,alpha);
        if( 0 ) { printf("   AA := "); polnsparseprint64s( A, X, na, n ); printf(";\n"); }
        if( 0 ) { printf("   Fz := "); taylorprint64s( dFz, alpha, FC, FX, nF, n-1 ); printf(";\n"); }
        if( 0 ) { printf("   Gz := "); taylorprint64s( dGz, alpha, GC, GX, nG, n-1 ); printf(";\n"); }
        if( 0 ) { printf("   beta := "); vecprint64s(beta,n); printf(";\n"); }
        printf("   Aval = %lld\n", Aval );
        printf("   Fval = %lld\n", Fval );
        printf("   Gval = %lld\n", Gval );
    }
    return( temp==0 );
}


void hensparseprint64s( LONG *A, LONG *X, int t, char *c, int n ) {
    int numbits,i,j,d;
    LONG m,mask,D[63];
    if( t==0 ) { printf("0"); return; }
    numbits = 63/n;
    mask = (1LL<<numbits)-1;
    for( i=0; i<t; i++ ) {
        m = X[i];
        printf("%lld",A[i]);
        for( j=0; j<n; j++ ) {
            D[j] = m & mask;
            m = m >> numbits;
        }
        for( j=1; j<=n; j++ ) {
            d = D[n-j];
            if( d!=0 ) printf("*%c^%d",c[j-1],d);
        }
        if( i!=t-1 ) printf("+");
    }
    return;
}


void solvecheck( LONG *m, LONG *a, LONG *b, int n, LONG p ) {
    int i, j; LONG c, *M; recint P;
    //   check sum( M[i]^(j+1) a[i], i=0..n-1 ) = b[j]  for  j=0..n-1
    P = recip1(p);
    M = array(n);
    for( i=0; i<n; i++ ) M[i] = 1;
    for( i=0; i<n; i++ ) {
       for( j=0; j<n; j++) M[j] = mulrec64(M[j],m[j],P);
       for( c=0,j=0; j<n; j++ ) c = add64s(c,mulrec64(M[j],a[j],P),p);
       if( c!=b[i] ) { printf("\n ** solve failure ** \n"); freemem(M); return; }
    }
    printf("solve is correct\n");
    freemem(M);
    return;
}

void polnsparsevecdeg64s( LONG *X, int t, int n, int *D );
LONG polbinsearch64s( LONG *A, LONG *X, int na, LONG mon );
void polSolve64s( LONG *R, int n, LONG *L, LONG *a, LONG *X, LONG p );
void polprint64s( LONG *A, int d );
void vecprint64s( LONG *A, int n );
void veccopy64s(LONG* A, int n, LONG* B);
int HornerExpand( int d, LONG alpha, int n, int *N, LONG **F, LONG **Y, LONG *A, LONG *X, LONG *T1, LONG *T2, LONG p );
int Expand( int d, int n, int *N, LONG **F, LONG **Y, LONG *A, LONG *X, LONG p );


int HenselLift2( LONG *A, LONG *X, int na, LONG *f, LONG *u, int nf, LONG *g, LONG *v, int ng, LONG alpha, LONG p,
                 int NumCores, //LONG *Ak, LONG *Xk, LONG *f0, LONG *g0,
                 int **DegFi, LONG ***Fxyi, int **DegGi, LONG ***Gxyi, // temporary space
                 int *DegFz, int *nF, LONG **F, LONG **Y, int *DegGz, int *nG, LONG **G, LONG **Z );


void polbivarimage64s( int dx, int dz, LONG ***F, int **D );

void getspaceMTSHLimages( int dy, int dx, int dz, int ***D, LONG ****F ) {
    // space for dy+1 images (of size (dx+1)(dz+1)) computed by HenselLift2
int i;
    D[0] = array2ints( dy+1, dz+2 ); // D[0][i][j+1] = degree(F[0][i][j],x)
    F[0] = (LONG ***) arrayofptrs( dy+1 );
    for( i=0; i<=dy; i++ ) F[0][i] = array2(dz+1,dx+1);
    return;
}


void VandermondeSolver( LONG *R, int n, LONG *LAMBDA, LONG *a, LONG *b, LONG p ) {
    LONG i,zero;
    recint P;
    P = recip1(p);
    // getlambda( R, n, LAMBDA, p );
    for( zero=1,i=0; i<n && zero; i++ ) if( b[i]!=0 ) zero = 0;
    if( zero ) { for( i=0; i<n; i++ ) a[i]=0; return; }
    treemul( R, n, LAMBDA, a, p );  // = product( x-R[i], i=0..n-1 )
    polSolve64s( R, n, LAMBDA, b, a, p ); // solve V a = b for a
    for( i=0; i<n; i++ ) if( a[i]!=0 ) a[i] = mulrec64(a[i],modinv64s(R[i],p),P); // V is shifted
    if( DEBUG ) solvecheck( R, a, b, n, p );
    return;
}

int VandermondeSolverCilk( LONG *R, int n, LONG *LAMBDA, LONG *a, LONG *b, LONG p ) {
    VandermondeSolver( R, n, LAMBDA, a, b, p );
    return 1;
}



# define FAIL 0
# define ERROR 1
# define GOOD 2


int MTSHL( int NumCores,
           LONG * A, LONG * X, int na, 
           LONG * f, LONG * Y, int nf, 
           LONG * g, LONG * Z, int ng,
           int j, LONG alpha,
           //LONG ** F, LONG ** U, int *nF,
           //LONG ** G, LONG ** V, int *nG,
           LONG p )
{
    int i,k,m,n,s,dfx,dfy,dfz,dgx,dgy,dgz,dax,day,daz,sfmax, sgmax, failures, *NF, *NG, sf, sg, D[32], nAk, nfk, ngk;
    LONG nA2, nf2, beta[32], *BETA[32], *AG, *AW, *fG, *fW, *gG, *gW, *Ak, *Xk, *fk, *Yk, *gk, *Zk, *LAMBDA, *a, *b, *R;
    int *nFk, *nGk, **nF, **nG, NumImages;
    LONG **Fk, **Uk, **Gk, **Vk, ***F, ***U, ***G, ***V , *aa, *xx;
    double ST,ET,TT,H2,HT,VT,IT,ITfg;
    extern double SORTTIME;
    recint P;
    int **ImageFDeg, **ImageGDeg; // space for Hensel2
    LONG ***ImageFCof, ***ImageGCof; // space for Hensel2
    LONG *AAk[64], *AAk2[64], *AXk[64], *AXk2[64];
    LONG *ffk[64], *ffk2[64], *fYk[64], *fYk2[64];

    if( NumCores<1 || NumCores>64 ) return ERROR;
    if( j<=3 || j>=32 ) return ERROR;
    if( p<3 ) return ERROR;
    P = recip1(p);
    n = j-3; // we evaluate out n variables x3,...,x3+n-1

    printf("MTSHL:  j=%d  n=%d  p=%lld  alpha=%lld  NumCores=%d\n",j,n,p,alpha,NumCores);
    TT = realtime();
    H2 = 0.0;
    SORTTIME = 0.0;

    if( 0 ) { printf(" A := "); polnsparseprint64s( A, X, na, j ); printf(";\n"); }
    if( 0 ) { printf(" f := "); polnsparseprint64s( f, Y, nf, j-1 ); printf(";\n"); }
    if( 0 ) { printf(" g := "); polnsparseprint64s( g, Z, ng, j-1 ); printf(";\n"); }

    polnsparsevecdeg64s(Y,nf,j-1,D); dfx = D[j-2]; dfy = D[j-3];
    polnsparsevecdeg64s(Z,ng,j-1,D); dgx = D[j-2]; dgy = D[j-3];
    polnsparsevecdeg64s(X,na,j,D); dax = D[j-1]; day = D[j-2]; daz = D[0];
       printf("deg(f,x1)=%d  deg(f,x2)=%d\n", dfx, dfy);
       printf("deg(g,x1)=%d  deg(g,x2)=%d\n", dgx, dgy);
       printf("deg(a,x1)=%d  deg(a,x2)=%d  deg(a,xj)=%d\n", dax, day, daz);

    if( dfx+dgx!=dax || dfy+dgy!=day ) return ERROR;
    // the degrees computed in D are backwards - so reverse them to fix bugs !
    for( i=0; i<j/2; i++ ) { k = D[i]; D[i] = D[j-1-i]; D[j-1-i] = k; }
    for( i=0; i<j; i++ ) printf(" deg(A,x%d)=%d ",i,D[i]); printf("\n"); 
    for( i=0,k=2; i<n; i++,k++ ) BETA[i] = array(D[k]+1);

    NF = arrayints((dfx+1)*(dgx+1)); sf = suppsizes(Y,nf,j-1,2,NF);
    NG = arrayints((dgx+1)*(dgy+1)); sg = suppsizes(Z,ng,j-1,2,NG);
    for( sfmax=0,i=0; i<sf; i++ ) sfmax = max32s( sfmax, NF[i] );
    for( sgmax=0,i=0; i<sg; i++ ) sgmax = max32s( sgmax, NG[i] );

    printf("MTSHL:  sfmax = %d  sgmax = %d\n", sfmax, sgmax );

    s = max32s( sfmax, sgmax ); // the number of images needed

    fG = array(nf); fW = array(nf);
    gG = array(ng); gW = array(ng);
    AG = array(na); AW = array(na);


    failures = 0;
    while( failures < 3 ) { // compute arrays of powers of beta's

printf(" MTSHL BETA \n" );
        for( i=0,k=2; i<n; i++,k++ ) {
           do { beta[i] = rand64s(p); } while ( beta[i]<2 );
           evalbeta(BETA[i],D[k],beta[i],P);
        }

    printf("MTSHL:  n=%d\n",n);
    for( i=0; i<n; i++ ) printf(" beta[%d]=%lld\n", i+3, beta[i] );

if( DEBUG ) {
    aa = array(na); veccopy64s(A,na,aa);
    xx = array(na); veccopy64s(X,na,xx);
}

    IT = 0.0;
    ST = realtime();

        printf(" MTSHL evalinit(f) \n" );
        evalinit( Y, nf, j-1, 3, j-1, BETA, fG, fW, p );  // j = 4
        if( !distinctsupp( Y, fG, nf, j-1, 2 ) ) { failures ++; printf("  supp(f) evals not distinct\n"); continue; }
        evalinit( Z, ng, j-1, 3, j-1, BETA, gG, gW, p );
        if( !distinctsupp( Z, gG, ng, j-1, 2 ) ) { failures ++; printf("  supp(g) evals not distinct\n"); continue; }


        //evalinit( X, na, j, 3, j-1, BETA, AG, AW, p );
             //if(1) {    printf(" Gam := "); vecprint64s( AG, na ); printf("; \n"); }
             //if(1) {    printf(" A := "); polnsparseprint64s( A, X, na, j ); printf(";\n"); }

    ITfg = realtime()-ST;
    SORTTIME=realtime();
        reorderxn( A, X, na, j );
        k =  cilksort64s( A, X, na );
        
        if( DEBUG) { for( k=0; k<na-1; k++ ) if( X[k]>X[k+1] ) printf("NOT SORTED IN HENSEL\n"); }
    SORTTIME = realtime()-SORTTIME;

        //evalinit( X, na, j, 4, j, BETA, AG, AW, p );
        //if(1) {    printf(" G := "); vecprint64s( AG, na ); printf("; \n"); }
        printf(" MTSHL supp(f) & supp(g) are distinct\n" );
        k =  cilkinitpar( NumCores, X, na, j, 4, j, BETA, AG, AW, p ); 
        

    IT = realtime()-ST-SORTTIME;
    printf(" INITTIME=%10.3f\n",IT);
    printf(" SORTTIME=%10.3f\n",SORTTIME);
    ET = 0.0;


//printf(" Ainit := "); polnsparseprint64s( A, AW, na, 3 ); printf(";\n");
//printf(" finit := "); polnsparseprint64s( f, fW, nf, 2 ); printf(";\n");
//printf(" ginit := "); polnsparseprint64s( g, gW, ng, 2 ); printf(";\n");

        break;
    }
    
    if( failures==3 ) { printf("MTSHL  FAILED\n"); return FAIL; }

    if( nf<50000 ) { fk = array(nf); Yk = array(nf); }
    else {
        for( i=0; i<NumCores; i++ ) {
            ffk[i] = array(nf); ffk2[i] = array(nf);
            fYk[i] = array(nf); fYk2[i] = array(nf); 
        }
        fk = ffk[0]; Yk = fYk[0];
    }

    gk = array(ng); Zk = array(ng);
    m = min32s( na, (dax+1)*(day+1)*(daz+1) );
    printf("MTSHL:  dax=%d  day=%d  daz=%d  m=%d\n",dax,day,daz,m);
 
    NumImages = 2; // do two images at a time
    for( i=0; i<NumCores; i++ ) { AAk[i] = array(m); AAk2[i] = array(m); AXk[i] = array(m); AXk2[i] = array(m); }
    Ak = AAk[0];
    Xk = AXk[0];

    getspaceMTSHLimages( day, dax, daz, &ImageFDeg, &ImageFCof );
    getspaceMTSHLimages( day, dax, daz, &ImageGDeg, &ImageGCof );

    nF = arrayptrsints(s);
    nG = arrayptrsints(s);
    F  = (LONG ***) arrayofptrs(s);
    U  = (LONG ***) arrayofptrs(s);
    G  = (LONG ***) arrayofptrs(s);
    V  = (LONG ***) arrayofptrs(s);

for( k=0; k<s; k++ ) {

    printf("MTSHL: image %d of %d  : ",k+1,s);

    ST = realtime();

    //nAk = evalnext( A, AW, AG, na, Ak, Xk, p ); //  A is updated
    //nAk =  cilkevalpar( A, AW, AG, na, AAk, AXk, NumCores, p ); //  A is updated 
    //
    //Ak = AAk[0];
    //Xk = AXk[0];


    //if( k%2==0 ) nA2 = evalnext2( A, AW, AG, na, AAk[0], AAk2[0], AXk[0], AXk2[0], p, m ); //  A is updated
    //printf(" unpack1=%d  unpack2=%d \n", unpacklong(nA2,1), unpacklong(nA2,2) );
    //if( k%2==0 ) { Ak = AAk[0]; Xk = AXk[0]; nAk = unpacklong(nA2,1); } 
    //else { Ak = AAk2[0]; Xk = AXk2[0]; nAk = unpacklong(nA2,2); }

    // Get the next two evaluations of A. NB: A is updated 
    if( k%2==0 ) { nA2 =  cilkevalpar2( A, AW, AG, na, AAk, AAk2, AXk, AXk2, NumCores,  p, m );  }
    if( k%2==0 ) { Ak = AAk[0]; Xk = AXk[0]; nAk = unpacklong(nA2,1); } 
    else         { Ak = AAk2[0]; Xk = AXk2[0]; nAk = unpacklong(nA2,2); }

    ET += realtime()-ST; // just count evaluation of a(x1,x2,beta^k,xj) for reporting purposes

if( nf<50000 ) nfk = evalnext( f, fW, fG, nf, fk, Yk, p ); //  f is updated
else { // evaluate f in parallel
    if( k%2==0 ) { nf2 =  cilkevalpar2( f, fW, fG, nf, ffk, ffk2, fYk, fYk2, NumCores,  p, m );  }
    if( k%2==0 ) { fk = ffk[0]; Yk = fYk[0]; nfk = unpacklong(nf2,1); }
    else { fk = ffk2[0]; Yk = fYk2[0]; nfk = unpacklong(nf2,2); }
}

    ngk = evalnext( g, gW, gG, ng, gk, Zk, p ); //  g is updated

    printf("MTSHL: #A=%d  #Ak=%d  #f=%d  #fk=%d  #g=%d  #gk=%d\n", na, nAk, nf, nfk, ng, ngk );

if( DEBUG ) {
if(0) {  printf(" A%d := ",k+1); polnsparseprint64s( Ak, Xk, nAk, 3 ); printf(";\n");  }
if(0) {  printf(" f%d := ",k+1); polnsparseprint64s( fk, Yk, nfk, 2 ); printf(";\n");  }
if(0) {  printf(" g%d := ",k+1); polnsparseprint64s( gk, Zk, ngk, 2 ); printf(";\n");  }
}

    //printf("MTSHL: deg(A,z)=%d\n", daz);
    nFk = arrayints(daz+1); // numterms(F[i])
    nGk = arrayints(daz+1); // numterms(G[i])
    Fk = array2(daz+1,(dfx+1)*(dfy+1));
    Uk = array2(daz+1,(dfx+1)*(dfy+1));
    Gk = array2(daz+1,(dfx+1)*(dfy+1));
    Vk = array2(daz+1,(dfx+1)*(dfy+1));

    ST = realtime();
    i =  HenselLift2( Ak, Xk, nAk, fk, Yk, nfk, gk, Zk, ngk, alpha, p, // inputs
                     NumCores, //H2Ak, H2Xk, H2f0, H2g0,
                     ImageFDeg, ImageFCof, ImageGDeg, ImageGCof, // temporary space
                     &dfz, nFk, Fk, Uk, &dgz, nGk, Gk, Vk );
    
    H2 += realtime()-ST;
    
    if( i==0 ) { printf("HenselLift2 failed!\n"); return 0; }

    //printf("DONE HenselLift2:  dfz = %d  dgz = %d\n", dfz, dgz );

    if( DEBUG )
    //if( henselcheck( Ak, Xk, nAk, 3, dfz, Fk, Uk, nFk, dgz, Gk, Vk, nGk, alpha, p ) ) 
    if( henselcheck( Ak, Xk, nAk, 3, dfz, Fk, Uk, nFk, dgz, Gk, Vk, nGk, 0, p ) ) 
         printf( "Hensel: image %d OKAY\n", k+1 );
    else printf( "Hensel: image %d FAIL\n", k+1 );


    // record the results in nFk, Fk, Uk, nGk, Gk, Vk
    nF[k] = nFk; F[k] = Fk; U[k] = Uk;
    nG[k] = nGk; G[k] = Gk; V[k] = Vk;


/*
    for( i=0; i<=dfz; i++ ) {
       if( i==0 ) printf( "Fk := "); else printf("+");
       printf("(");
       //polnsparseprint64s(Fk[i],Uk[i],nFk[i],2);
       hensparseprint64s(Fk[i],Uk[i],nFk[i],"yx",2);
       printf(")*(z-%lld)^%d ",alpha,i);
    }
    printf(";\n");
    for( i=0; i<=dgz; i++ ) {
       if( i==0 ) printf( "Gk := "); else printf("+");
       printf("(");
       //polnsparseprint64s(Gk[i],Vk[i],nGk[i],2);
       hensparseprint64s(Gk[i],Vk[i],nGk[i],"yx",2);
       printf(")*(z-%lld)^%d",alpha,i);
    }
    printf(";\n");
*/

}

{    LONG x, mask, mon, dy, dx;
     int m, numbits,ss,tt,nn, *FN, *GN, dummy, nmax;
     LONG **FX, **FC, **GX, **GC;
     LONG *amem, *bmem, *Lmem;

     numbits=63/(j-1);             
     mask = (1LL << numbits) - 1;

printf("SOLVE:  s = %d\n",s);

     VT = realtime();

     //a = array(s);
     //b = array(s);
     //LAMBDA = array(s+1);
     //R = array(s);

     FN = arrayints(dfz+1); // space for final sigmas
     FC = array2(dfz+1,nf); //FC = arrayofptrs(dfz+1); for( i=0; i<=dfz; i++ ) FC[i] = array(nf);
     FX = array2(dfz+1,nf); //FX = arrayofptrs(dfz+1); for( i=0; i<=dfz; i++ ) FX[i] = array(nf);

     printf("\nF: system sizes = "); vecprint32s(NF,sf);

     ss = 0; for (m=0; m<sf; m++ ) ss += NF[m];
     amem = array(ss); bmem = array(ss); Lmem = array(ss+sf);

     for( k=0; k<=dfz; k++ ) {

/*
         a = amem; b = bmem; LAMBDA = Lmem;
         for( tt=m=0; m<sf; m++,tt+=ss ) {
              ss = NF[m];
              //x = Y[tt] >> (j-3)*numbits;
              x = Y[tt] >> (n*numbits);
              dx = x & mask;
              dy = x >> numbits;
              mon = (dx<<31)|dy;
              R = fG+tt; //for( i=0; i<ss; i++ ) R[i] = fG[tt+i];
              for( i=0; i<ss; i++ ) b[i] = polbinsearch64s(F[i][k],U[i][k],nF[i][k],mon);
              if( DEBUG ) printf("Solve F(%d,%d): s=%d ... ",k,m,ss);
              if( ss<100 ) VandermondeSolver( R, ss, LAMBDA, a, b, p );
              else dummy =  VandermondeSolverCilk( R, ss, LAMBDA, a, b, p );
              a += ss; b += ss; LAMBDA += ss+1;
         }
*/

         // MBM: I've separated the solving process into 4 steps
         // 1: Get the right-hand-side vectors -- the b's
         // 2: Start solving the larger systems in parallel
         // 3: Now solve the smaller systems on the master thread
         // 4: Extract the solutions
         // MBM: Then repeat this for the other factor g

         b = bmem;
         for( tt=m=0; m<sf; m++,tt+=ss ) {
              ss = NF[m];
              x = Y[tt] >> (n*numbits); //x = Y[tt] >> (j-3)*numbits;
              dx = x & mask;
              dy = x >> numbits;
              mon = (dx<<31)|dy;
              for( i=0; i<ss; i++ ) b[i] = polbinsearch64s(F[i][k],U[i][k],nF[i][k],mon);
              b += ss;
         }

         a = amem; b = bmem; LAMBDA = Lmem; R = fG;
         for( m=0; m<sf; m++ ) { // solve the large ones in parallel
              ss = NF[m];
              if( ss>=25 && DEBUG ) printf("Solve F(%d,%d): s=%d ... ",k,m,ss);
              if( ss>=25 ) dummy =  VandermondeSolverCilk( R, ss, LAMBDA, a, b, p );
              R += ss; a += ss; b += ss; LAMBDA += (ss+1);
         }

         a = amem; b = bmem; LAMBDA = Lmem; R = fG;
         for( m=0; m<sf; m++ ) {  // solve the small ones serially
              ss = NF[m];
              if( ss<25 && DEBUG ) printf("Solve F(%d,%d): s=%d ... ",k,m,ss);
              if( ss<25 ) VandermondeSolver( R, ss, LAMBDA, a, b, p );
              R += ss; a += ss; b += ss; LAMBDA += (ss+1);
         }

          // wait for everyone to finish

         a = amem;
         nn = 0;
         for( tt=m=0; m<sf; m++,tt+=ss ) { // extract solutions 
              ss = NF[m];
              for( i=0; i<ss; i++ ) // record answer
                  if( a[i]!=0 ) { FC[k][nn] = a[i]; FX[k][nn] = Y[tt+i]; nn++; }
              a += ss;
         }
         FN[k] = nn;
     }

     GN = arrayints(dgz+1);  // space for final taus
     GC = array2(dgz+1,ng); //GC = arrayofptrs(dgz+1); for( i=0; i<=dgz; i++ ) GC[i] = array(ng);
     GX = array2(dgz+1,ng); //GX = arrayofptrs(dgz+1); for( i=0; i<=dgz; i++ ) GX[i] = array(ng);

     printf("\n\nG: system sizes = "); vecprint32s(NG,sg); printf("\n\n");

     freemem(amem); freemem(bmem); freemem(Lmem);
     ss = 0; for (m=0; m<sg; m++ ) ss += NG[m];
     amem = array(ss); bmem = array(ss); Lmem = array(ss+sg);

     for( k=0; k<=dgz; k++ ) {

/*
         a = amem; b = bmem; LAMBDA = Lmem;
         for( tt=m=0; m<sg; m++,tt+=ss ) {
              ss = NG[m];
              //x = Z[tt] >> (j-3)*numbits;
              x = Z[tt] >> n*numbits;
              dx = x & mask;
              dy = x >> numbits;
              mon = (dx<<31)|dy;
              R = gG+tt; //for( i=0; i<ss; i++ ) R[i] = gG[tt+i];
              for( i=0; i<ss; i++ ) b[i] = polbinsearch64s(G[i][k],V[i][k],nG[i][k],mon);
              if( DEBUG ) printf("Solve G(%d,%d): s=%d ... ",k,m,ss);
              if( ss<100 ) VandermondeSolver( R, ss, LAMBDA, a, b, p );
              else dummy =  VandermondeSolverCilk( R, ss, LAMBDA, a, b, p );
              a += ss; b += ss; LAMBDA += (ss+1);
         }
*/

         b = bmem;
         for( tt=m=0; m<sg; m++,tt+=ss ) {
              ss = NG[m];
              //x = Z[tt] >> (j-3)*numbits;
              x = Z[tt] >> n*numbits;
              dx = x & mask;
              dy = x >> numbits;
              mon = (dx<<31)|dy;
              for( i=0; i<ss; i++ ) b[i] = polbinsearch64s(G[i][k],V[i][k],nG[i][k],mon);
              b += ss;
         }

         a = amem; b = bmem; LAMBDA = Lmem; R = gG;
         for( m=0; m<sg; m++ ) { // solve the large ones in parallel
              ss = NG[m];
              if( ss>=25 ) if( DEBUG ) printf("Solve G(%d,%d): s=%d ... ",k,m,ss);
              if( ss>=25 ) dummy =  VandermondeSolverCilk( R, ss, LAMBDA, a, b, p );
              R += ss; a += ss; b += ss; LAMBDA += (ss+1);
         }

         a = amem; b = bmem; LAMBDA = Lmem; R = gG;
         for( m=0; m<sg; m++ ) {  // solve the small ones serially
              ss = NG[m];
              if( ss<25 ) if( DEBUG ) printf("Solve G(%d,%d): s=%d ... ",k,m,ss);
              if( ss<25 ) VandermondeSolver( R, ss, LAMBDA, a, b, p );
              R += ss; a += ss; b += ss; LAMBDA += (ss+1);
         }

         

         a = amem;
         nn = 0;
         for( tt=m=0; m<sg; m++,tt+=ss ) { // extract solutions 
              ss = NG[m];
              for( i=0; i<ss; i++ ) // record answer
                  if( a[i]!=0 ) { GC[k][nn] = a[i]; GX[k][nn] = Z[tt+i]; nn++; }
              a += ss;
         }
         GN[k] = nn;
     }
     VT = realtime()-VT;

     //freemem(a); freemem(b); freemem(LAMBDA); freemem(R);
     freemem(amem); freemem(bmem); freemem(Lmem);

     printf("MTSHL:  solve memory freed\n");


if( DEBUG ) {
    //if( henselcheck( aa, xx, na, j, dfz, FC, FX, FN, dgz, GC, GX, GN, alpha, p ) ) 
    if( henselcheck( aa, xx, na, j, dfz, FC, FX, FN, dgz, GC, GX, GN, 0, p ) )
         printf( "Hensel lift check OKAY\n" );
    else printf( "Hensel lift check FAIL\n" );
    freemem(aa);
    freemem(xx);
}

printf("FN = "); vecprint32s(FN,dfz+1); printf("\n");
printf("GN = "); vecprint32s(GN,dgz+1); printf("\n");
    
{   LONG *FFC, *FFX, *GGC, *GGX; int nf, ng, nt;
    ST = realtime();
    //for( nFmax=0,i=0; i<=dfz; i++ ) if( FN[i]>nFmax ) nFmax = FN[i];
    //for( nGmax=0,i=0; i<=dgz; i++ ) if( GN[i]>nGmax ) nGmax = GN[i];
    //printf("nf=%d  ng=%d\n",nf,ng);
    //FFC = array( (dfz+1)*FN[0] );
    //FFX = array( (dfz+1)*FN[0] );
    //GGC = array( (dgz+1)*GN[0] );
    //GGX = array( (dgz+1)*GN[0] );
    nf = 0; for( i=0; i<=dfz; i++ ) if( FN[i]>=0 ) nf += FN[i];
    FFC = array(nf);
    FFX = array(nf);
    ng = 0; for( i=0; i<=dgz; i++ ) if( GN[i]>=0 ) ng += GN[i];
    GGC = array(ng);
    GGX = array(ng);
    //if( nFmax>nGmax ) nt = nFmax*(dfz+1); else nt = nGmax*(dfz+1);
    //printf("nt = %lld  nFmax=%lld  nGmax=%lld\n",nt,nFmax,nGmax);
    //T1 = array(nt); T2 = array(nt);
if( DEBUG ) printf("Horner Expand of F starting\n");
    //nf = HornerExpand( dfz, alpha, j-1, FN, FC, FX, FFC, FFX, T1, T2, p );
    //nf = HornerExpand( dfz, 0, j-1, FN, FC, FX, FFC, FFX, T1, T2, p );
    nf = Expand( dfz, j-1, FN, FC, FX, FFC, FFX, p );
    //printf("  FF := "); polnsparseprint64s(FFC, FFX, nf, j); printf(";\n");
if( DEBUG ) printf("Horner Expand of G starting\n");
    //ng = HornerExpand( dgz, alpha, j-1, GN, GC, GX, GGC, GGX, T1, T2, p );
    //ng = HornerExpand( dgz, 0, j-1, GN, GC, GX, GGC, GGX, T1, T2, p );
    ng = Expand( dgz, j-1, GN, GC, GX, GGC, GGX, p );
    //printf("  GG := "); polnsparseprint64s(GGC, GGX, ng, j); printf(";\n");
    //freemem(T1);
    //freemem(T2);
    HT = realtime()-ST;
if( DEBUG ) printf("Horner Expand done\n");
}

}

    TT = realtime()-TT;
    printf("MTSHL(s=%d) Time=%9.3fs  Hen2=%8.3fs  Sort=%7.3f  Eval=%8.3f  Solve=%7.3f  Horner=%7.3f  Init=%7.3f  Initfg=%7.3f\n",s,TT,H2,SORTTIME, ET,VT, HT, IT, ITfg);


    for( i=0; i<n; i++ ) freemem(BETA[i]);

if( nf<50000 ) { freemem(fk); freemem(Yk); }
    freemem(gk); freemem(Zk);
    freemem(NF); freemem(NG);
    freemem(fG); freemem(fW);
    freemem(gG); freemem(gW);
    freemem(AG); freemem(AW);

    for( i=0; i<NumCores; i++ ) { 
        freemem(AAk[i]);
        freemem(AAk2[i]);
        freemem(AXk[i]);
        freemem(AXk2[i]); }

if( nf>=50000 ) {
    for( i=0; i<NumCores; i++ ) { 
        freemem(ffk[i]);
        freemem(ffk2[i]);
        freemem(fYk[i]);
        freemem(fYk2[i]); }
}
    
    for( k=0; k<s; k++ ) { LONG *T;
        nFk = nF[k]; freemem(nFk); 
        Fk = F[k]; T = Fk[0]; freemem(T); freemem(Fk); 
        Uk = U[k]; T = Uk[0]; freemem(T); freemem(Uk);
        nGk = nG[k]; freemem(nGk);
        Gk = G[k]; T = Gk[0]; freemem(T); freemem(Gk); 
        Vk = V[k]; T = Vk[0]; freemem(T); freemem(Vk);
    }
    freemem(nF); freemem(F); freemem(U);
    freemem(nG); freemem(G); freemem(V);

    for( i=0; i<=day; i++ ) { LONG *T;
        T = ImageFCof[i][0]; freemem(T); freemem(ImageFCof[i]);
        T = ImageGCof[i][0]; freemem(T); freemem(ImageGCof[i]);
    }
    freemem(ImageFCof);    freemem(ImageGCof);
    freemem(ImageFDeg[0]); freemem(ImageFDeg);
    freemem(ImageGDeg[0]); freemem(ImageGDeg);
    //                 ImageFDeg, ImageFCof, ImageGDeg, ImageGCof, // temporary space
        
    printf("MTSHL: memory cleanup done\n");
    return GOOD;

}
