#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define LONG long long int
#define ULONG unsigned long long int
#define INT128 __int128_t
#define UINT128 __uint128_t

LONG * array( int n );
ULONG * arrayu( int n );
LONG ** arrayofptrs( int n );
char * arrayofbytes( int n );
void vecprint64u(ULONG*A, int n);
void randpolyNOTmonic64Z( int n, int *DEG, int D, int t, LONG *A, LONG *X, LONG p, int m );

/******************************************************************************************/
/*  Zp utilities                                                                          */
/******************************************************************************************/

#define modinv64s modinv64szutil

int abs32s(int x){return (x<0)?(-x):x;}
LONG add64s(LONG a, LONG b, LONG p);
LONG sub64s(LONG a, LONG b, LONG p);
int max32s(int a, int b) { if(a>b) return a; else return b; }
int min32s(int a, int b) { if(a<b) return a; else return b; }
LONG add64s(LONG a, LONG b, LONG p) { LONG t; t = (a-p)+b; t += (t>>63) & p; return t; }
LONG sub64s(LONG a, LONG b, LONG p) { LONG t; t = a-b; t += (t>>63) & p; return t; }
LONG neg64s(LONG a, LONG p) { LONG t; t = 0-a; t += (t>>63) & p; return t; }
LONG mul64s(LONG a, LONG b, LONG p) {
        LONG q, r;
        __asm__ __volatile__(           \
        "       mulq    %%rdx           \n\t" \
        "       divq    %4              \n\t" \
        : "=a"(q), "=d"(r) : "0"(a), "1"(b), "rm"(p));
        return r;
}

/* c^(-1) mod p assuming 0 < c < p < 2^63 */
LONG modinv64s( LONG c, LONG p )
{   LONG d,r,q,r1,c1,d1;
    d = p;
    c1 = 1;
    d1 = 0;
    while( d != 0 ) {
        q = c / d;
        r = c - q*d; r1 = c1 - q*d1;
        c = d; c1 = d1;
        d = r; d1 = r1;
    }
    if( c!=1 ) return( 0 );
    if( c1 < 0 ) c1 += p;
    return( c1 );
}

#define UINT64 unsigned long long
typedef struct {
	UINT64 s;	/* shift */
	UINT64 v;	/* reciprocal of d */
	UINT64 d0;	/* divisor shifted up */
	UINT64 d1;
} recint;

UINT64 mulrec64(UINT64  a, UINT64  b, recint  v);
recint recip1(UINT64  p);			/* reciprocal of  64-bit p */
void fma128(UINT64 *x, UINT64  a, UINT64  b);
void add128(UINT64 *x, UINT64 *y);


LONG zmodp( ULONG *A, int n, ULONG p ) {
// input a = A[0]+A[1]*B + ... + A[n-1]*B^(n-1) where B=2^64 
// output a mod p
    LONG x,y;
    ULONG B;
    for( n--; n>=0 && A[n]==0; n-- );
    if( n<0 ) return 0;
    B = 18446744073709551615llu;
    x = B%p;
    x = add64s(x,1,p);
    y=A[n]%p;

    for( n--; n>=0; n-- ) y = add64s( mul64s(y,x,p), A[n]%p, p );
    return y;
}



LONG zmodP64acc( ULONG *A, int n, LONG p, recint P,  LONG *X ) {
// input a = A[0]+A[1]*B + ... + A[n-1]*B^(n-1) where B=2^64 
// output a mod p
    for( n--; n>=0 && A[n]==0; n-- );
    if( n<0 ) return 0;
    INT128 y; INT128 t;int i;
    INT128 s = p; s = s<<64; 

    y = A[0];
    for (i = 1; i<=n;i++ ){ 
        y -= (~(y>>127)) & s; 
        t = (INT128)A[i]*X[i-1];
        y +=t;
    }
    y -= (~(y>>127)) & s; 
    y %= p; y += (y>>127)&p;
    return (LONG)y;
}

LONG zmodP( ULONG *A, int n, ULONG p, recint P, LONG x ) {
// input a = A[0]+A[1]*B + ... + A[n-1]*B^(n-1) where B=2^64 
// output a mod p
    LONG y; ULONG B;
    for( n--; n>=0 && A[n]==0; n-- );
    if( n<0 ) return 0;
    y = mulrec64(A[n],1,P);
    for( n--; n>=0; n-- ) y = add64s( mulrec64(y,x,P), mulrec64(A[n],1,P), p );
    return y;
}
    

void zacc64u( ULONG *A, int n, ULONG *B, int m )
// compute A += B base 2^64 
// A must be long enough to hold the sum which means n >= m
{   int i,carry;
    for( carry=0,i=0; i<m; i++ ) {
        A[i] = A[i] + carry;
        carry = (A[i]==0 && carry); // assumes (ULONG)(2^64-1) + 1 == 0
        A[i] += B[i];
        carry = carry | (A[i]<B[i]);
    }
    while( i<n && carry ) { A[i] = A[i] + carry; carry = (A[i]==0); i++; }
    return;
}

int zadd64u( ULONG *A, int n, ULONG *B, int m, ULONG *C )
// compute C = A+B base 2^64 and return the length of the sum
// C must be large enough to hold the sum
{   int i,carry,maxlen;
    maxlen = max32s(n,m);
    for( i=0; i<n; i++ ) C[i] = A[i];
    while( i<=maxlen ) { C[i] = 0; i++; }
    zacc64u( C, maxlen+1, B, m );
    for( i=maxlen; i>=0 && C[i]==0; i-- );
    return i+1;
}

int zmul64u( ULONG *A, int n, ULONG *B, int m, ULONG *C )
// compute C = A*B base 2^64 and return the length of the product
// C must be of size m+n to hold the product.
{   
    int i,j;
    ULONG x[2],y[2],carry;
    for( i=0; i<n+m; i++ ) C[i] = 0;
    while( n>0 && A[n-1]==0 ) n--;
    while( m>0 && B[m-1]==0 ) m--;
    if( m==0 || n==0 ) return 0;
    for( i=0; i<n; i++ ) {
        carry = 0;
        for( j=0; j<m; j++ ) {
            x[0] = carry; x[1] = 0;  //printf("carry := %llu; a := %llu; b := %llu; ",carry,A[i],B[j]);
            fma128(x,A[i],B[j]);     //printf("x[0] := %llu; x[1] := %llu;\n",x[0],x[1]);
            y[0] = C[i+j]; y[1] = 0;
            add128(x,y);             //printf("c := %llu; x[0] := %llu; x[1] := %llu; \n",y[0],x[0],x[1]);
            C[i+j] = x[0]; 
            carry = x[1];
        }
        C[i+j] = carry;
    }
    for( i=n+m; i>0 && C[i-1]==0; i-- );
    return i; // length of A*B
}

//compare |A| vs. |B|
//m, n: size of A and B
//return -1: |A| < |B|
//return  1: |A| > |B|
//return  0: |A| = |B|
int zcompabs(ULONG*A, ULONG*B, int m, int n){
    m--; while(m>=0 && A[m] == 0)m--;
    n--; while(n>=0 && B[n] == 0)n--;

    if (m > n){ return 1; } 
    if (m < n){ return -1; } 
    if (m == -1){ return 0; }  //both zeros

    while(m>=0 && A[m] == B[m])m--; 
    if (m == -1){ return 0; } 
    if (A[m]< B[m]){ return -1; } 
    return 1;
}

//A and B are positive MP numbers
//assume: A>=B
//result = A - B
void zsub64u(ULONG* result, ULONG* A, ULONG* B, int m, int n){
    while(m>0 && A[m-1]==0)m--; while(n>0 && B[n-1]==0)n--; //m>=n at this point
    int i = 0;
    INT128 tmp = 1; tmp = tmp<<64; tmp--; ULONG BASE = (ULONG)tmp;//2^64 - 1
    //value
    int borrow = 0;
    for (i = 0;i< n; i++){
        if (A[i]>=B[i]){
            result[i] = A[i]-B[i]; 
            if (result[i] == 0 && borrow){ result[i] = BASE;  //2^64 - 1;
            }else{ result[i] -= borrow; borrow = 0; } 
        }else{
            result[i] = BASE - (B[i]-A[i]) +1;  //-(B[i] +A[i]) + 2^64
            if (borrow){ result[i]--; } 
            borrow = 1;
        } 
    }
    for (i = n;i< m; i++){
        if (A[i] == 0 && borrow){ result[i] = BASE; } 
        else{ result[i] = A[i] - borrow; borrow = 0; }
    }
}




// A += B 
// sa, sb: signs 0 for +, 1 for -
// m>=n: A must be long enough to store A+=B
// return the sign of the accumulation
// answer stored in A
int szacc64u(ULONG*A, int sa, ULONG*B, int sb,int m, int n){
    if (n == 0){ return sa; } 
    if (sa == sb){ zacc64u(A,m, B, n); return sa; } 
    int comp = zcompabs(A, B, m, n);
    if (comp == 0){ for(int i = 0;i< m; i++){ A[i] = 0; } return 0; } 
    if (comp < 0){ zsub64u(A, B, A, n, m); return sb; } 
    zsub64u(A, A, B, m, n);
    return sa;
}



////////////////////////////////// testing utilities

ULONG seed, mult;
ULONG rand64u() {
    ULONG x,y;
    extern ULONG seed, mult;
    seed = mult*seed;
    x = seed >> 32;
    seed = mult*seed;
    y = seed >> 32;
    x = (x<<32) | y;
    return(x);
}

ULONG * randvec64u(int n) {
    int i; 
    ULONG * A;
    A = array(n);
    for( i=0; i<n; i++ ) A[i] = rand64u();
    return A;
}

void randvecbin(char* A, int n ){ for (int i = 0;i< n; i++){ A[i] =(char)(rand64u()% 2); } }

//the only difference from zprintu is that
//negative is denoted by 1
void zprintu2( LONG *A, int n, int sign )
{   int i;
    for( n--; n>=0 && A[n]==0; n-- );
    if( n<0 ) { printf("0"); return; }
    if( sign==1 ) printf("-(");
    printf("%llu",A[0]);
    for( i=1; i<=n; i++ ) printf("+%llu*2^%d",A[i],i*64);
    if( sign==1) printf(")");
    return;
}

//assume all positive
void veczprintu(LONG*A, int n, int m){
    //printf("{");
    printf("[");
    for (int i = 0;i< n; i++, A+=m){
        printf("("); zprintu2(A, m, 0); printf(")");
        if (i<n-1){printf(", "); } 
    }
    printf("]:\n");
    //printf("}:\n");
}



void zprintu( LONG *A, int n, int sign ) {   int i;
    for( n--; n>=0 && A[n]==0; n-- );
    if( n<0 ) { printf("0"); return; }
    if( sign==-1 ) printf("-(");
    printf("%llu",A[0]);
    for( i=1; i<=n; i++ ) printf("+%llu*2^%d",A[i],i*64);
    if( sign==-1) printf(")");
    return;
}

void print128s( INT128 a )
{
   ULONG * x;
   if( a<0 ) { printf("-"); a = -a; }
   x = (ULONG *) (&a);
   if( x[1]==0 ) printf("%llu",x[0]); else printf("%llu*2^64+%llu",x[1],x[0]);
}

